# Autogenerated by nbdev

d = { 'settings': { 'branch': 'main',
                'doc_baseurl': '/',
                'doc_host': 'https://fastmonai.no',
                'git_url': 'https://github.com/MMIV-ML/fastMONAI',
                'lib_path': 'fastMONAI'},
  'syms': { 'fastMONAI.dataset_info': { 'fastMONAI.dataset_info.MedDataset': ('dataset_info.html#meddataset', 'fastMONAI/dataset_info.py'),
                                        'fastMONAI.dataset_info.MedDataset.__init__': ( 'dataset_info.html#meddataset.__init__',
                                                                                        'fastMONAI/dataset_info.py'),
                                        'fastMONAI.dataset_info.MedDataset._auto_cache_path': ( 'dataset_info.html#meddataset._auto_cache_path',
                                                                                                'fastMONAI/dataset_info.py'),
                                        'fastMONAI.dataset_info.MedDataset._create_data_frame': ( 'dataset_info.html#meddataset._create_data_frame',
                                                                                                  'fastMONAI/dataset_info.py'),
                                        'fastMONAI.dataset_info.MedDataset._get_data_info': ( 'dataset_info.html#meddataset._get_data_info',
                                                                                              'fastMONAI/dataset_info.py'),
                                        'fastMONAI.dataset_info.MedDataset._load_cache': ( 'dataset_info.html#meddataset._load_cache',
                                                                                           'fastMONAI/dataset_info.py'),
                                        'fastMONAI.dataset_info.MedDataset._process_with_cache': ( 'dataset_info.html#meddataset._process_with_cache',
                                                                                                   'fastMONAI/dataset_info.py'),
                                        'fastMONAI.dataset_info.MedDataset._save_cache': ( 'dataset_info.html#meddataset._save_cache',
                                                                                           'fastMONAI/dataset_info.py'),
                                        'fastMONAI.dataset_info.MedDataset._visualize_single_case': ( 'dataset_info.html#meddataset._visualize_single_case',
                                                                                                      'fastMONAI/dataset_info.py'),
                                        'fastMONAI.dataset_info.MedDataset.calculate_target_size': ( 'dataset_info.html#meddataset.calculate_target_size',
                                                                                                     'fastMONAI/dataset_info.py'),
                                        'fastMONAI.dataset_info.MedDataset.fingerprint': ( 'dataset_info.html#meddataset.fingerprint',
                                                                                           'fastMONAI/dataset_info.py'),
                                        'fastMONAI.dataset_info.MedDataset.get_size_statistics': ( 'dataset_info.html#meddataset.get_size_statistics',
                                                                                                   'fastMONAI/dataset_info.py'),
                                        'fastMONAI.dataset_info.MedDataset.get_suggestion': ( 'dataset_info.html#meddataset.get_suggestion',
                                                                                              'fastMONAI/dataset_info.py'),
                                        'fastMONAI.dataset_info.MedDataset.get_volume_summary': ( 'dataset_info.html#meddataset.get_volume_summary',
                                                                                                  'fastMONAI/dataset_info.py'),
                                        'fastMONAI.dataset_info.MedDataset.summary': ( 'dataset_info.html#meddataset.summary',
                                                                                       'fastMONAI/dataset_info.py'),
                                        'fastMONAI.dataset_info.MedDataset.visualize_cases': ( 'dataset_info.html#meddataset.visualize_cases',
                                                                                               'fastMONAI/dataset_info.py'),
                                        'fastMONAI.dataset_info.get_class_weights': ( 'dataset_info.html#get_class_weights',
                                                                                      'fastMONAI/dataset_info.py'),
                                        'fastMONAI.dataset_info.preprocess_dataset': ( 'dataset_info.html#preprocess_dataset',
                                                                                       'fastMONAI/dataset_info.py'),
                                        'fastMONAI.dataset_info.suggest_patch_size': ( 'dataset_info.html#suggest_patch_size',
                                                                                       'fastMONAI/dataset_info.py')},
            'fastMONAI.external_data': { 'fastMONAI.external_data.MURLs': ('external_data.html#murls', 'fastMONAI/external_data.py'),
                                         'fastMONAI.external_data._create_nodule_df': ( 'external_data.html#_create_nodule_df',
                                                                                        'fastMONAI/external_data.py'),
                                         'fastMONAI.external_data._create_spine_df': ( 'external_data.html#_create_spine_df',
                                                                                       'fastMONAI/external_data.py'),
                                         'fastMONAI.external_data._dataset_available_locally': ( 'external_data.html#_dataset_available_locally',
                                                                                                 'fastMONAI/external_data.py'),
                                         'fastMONAI.external_data._df_sort_and_add_columns': ( 'external_data.html#_df_sort_and_add_columns',
                                                                                               'fastMONAI/external_data.py'),
                                         'fastMONAI.external_data._process_ixi_xls': ( 'external_data.html#_process_ixi_xls',
                                                                                       'fastMONAI/external_data.py'),
                                         'fastMONAI.external_data._process_medmnist_img': ( 'external_data.html#_process_medmnist_img',
                                                                                            'fastMONAI/external_data.py'),
                                         'fastMONAI.external_data.download_example_endometrial_cancer_data': ( 'external_data.html#download_example_endometrial_cancer_data',
                                                                                                               'fastMONAI/external_data.py'),
                                         'fastMONAI.external_data.download_example_spine_data': ( 'external_data.html#download_example_spine_data',
                                                                                                  'fastMONAI/external_data.py'),
                                         'fastMONAI.external_data.download_ixi_data': ( 'external_data.html#download_ixi_data',
                                                                                        'fastMONAI/external_data.py'),
                                         'fastMONAI.external_data.download_ixi_tiny': ( 'external_data.html#download_ixi_tiny',
                                                                                        'fastMONAI/external_data.py'),
                                         'fastMONAI.external_data.download_medmnist3d_dataset': ( 'external_data.html#download_medmnist3d_dataset',
                                                                                                  'fastMONAI/external_data.py'),
                                         'fastMONAI.external_data.download_spine_test_data': ( 'external_data.html#download_spine_test_data',
                                                                                               'fastMONAI/external_data.py')},
            'fastMONAI.utils': { 'fastMONAI.utils.MLflowUIManager': ('utils.html#mlflowuimanager', 'fastMONAI/utils.py'),
                                 'fastMONAI.utils.MLflowUIManager.__init__': ('utils.html#mlflowuimanager.__init__', 'fastMONAI/utils.py'),
                                 'fastMONAI.utils.MLflowUIManager._find_running_mlflow': ( 'utils.html#mlflowuimanager._find_running_mlflow',
                                                                                           'fastMONAI/utils.py'),
                                 'fastMONAI.utils.MLflowUIManager.check_mlflow_installed': ( 'utils.html#mlflowuimanager.check_mlflow_installed',
                                                                                             'fastMONAI/utils.py'),
                                 'fastMONAI.utils.MLflowUIManager.find_available_port': ( 'utils.html#mlflowuimanager.find_available_port',
                                                                                          'fastMONAI/utils.py'),
                                 'fastMONAI.utils.MLflowUIManager.is_mlflow_running': ( 'utils.html#mlflowuimanager.is_mlflow_running',
                                                                                        'fastMONAI/utils.py'),
                                 'fastMONAI.utils.MLflowUIManager.is_port_available': ( 'utils.html#mlflowuimanager.is_port_available',
                                                                                        'fastMONAI/utils.py'),
                                 'fastMONAI.utils.MLflowUIManager.start_ui': ('utils.html#mlflowuimanager.start_ui', 'fastMONAI/utils.py'),
                                 'fastMONAI.utils.MLflowUIManager.status': ('utils.html#mlflowuimanager.status', 'fastMONAI/utils.py'),
                                 'fastMONAI.utils.MLflowUIManager.stop': ('utils.html#mlflowuimanager.stop', 'fastMONAI/utils.py'),
                                 'fastMONAI.utils.ModelTrackingCallback': ('utils.html#modeltrackingcallback', 'fastMONAI/utils.py'),
                                 'fastMONAI.utils.ModelTrackingCallback.__init__': ( 'utils.html#modeltrackingcallback.__init__',
                                                                                     'fastMONAI/utils.py'),
                                 'fastMONAI.utils.ModelTrackingCallback._build_config': ( 'utils.html#modeltrackingcallback._build_config',
                                                                                          'fastMONAI/utils.py'),
                                 'fastMONAI.utils.ModelTrackingCallback._extract_epoch_metrics': ( 'utils.html#modeltrackingcallback._extract_epoch_metrics',
                                                                                                   'fastMONAI/utils.py'),
                                 'fastMONAI.utils.ModelTrackingCallback._extract_training_params': ( 'utils.html#modeltrackingcallback._extract_training_params',
                                                                                                     'fastMONAI/utils.py'),
                                 'fastMONAI.utils.ModelTrackingCallback._log_split_df': ( 'utils.html#modeltrackingcallback._log_split_df',
                                                                                          'fastMONAI/utils.py'),
                                 'fastMONAI.utils.ModelTrackingCallback._register_pytorch_model': ( 'utils.html#modeltrackingcallback._register_pytorch_model',
                                                                                                    'fastMONAI/utils.py'),
                                 'fastMONAI.utils.ModelTrackingCallback._save_model_artifacts': ( 'utils.html#modeltrackingcallback._save_model_artifacts',
                                                                                                  'fastMONAI/utils.py'),
                                 'fastMONAI.utils.ModelTrackingCallback.after_epoch': ( 'utils.html#modeltrackingcallback.after_epoch',
                                                                                        'fastMONAI/utils.py'),
                                 'fastMONAI.utils.ModelTrackingCallback.after_fit': ( 'utils.html#modeltrackingcallback.after_fit',
                                                                                      'fastMONAI/utils.py'),
                                 'fastMONAI.utils.ModelTrackingCallback.before_fit': ( 'utils.html#modeltrackingcallback.before_fit',
                                                                                       'fastMONAI/utils.py'),
                                 'fastMONAI.utils.ModelTrackingCallback.extract_all_params': ( 'utils.html#modeltrackingcallback.extract_all_params',
                                                                                               'fastMONAI/utils.py'),
                                 'fastMONAI.utils.ModelTrackingCallback.log_dataframe': ( 'utils.html#modeltrackingcallback.log_dataframe',
                                                                                          'fastMONAI/utils.py'),
                                 'fastMONAI.utils.ModelTrackingCallback.log_metrics': ( 'utils.html#modeltrackingcallback.log_metrics',
                                                                                        'fastMONAI/utils.py'),
                                 'fastMONAI.utils.ModelTrackingCallback.log_metrics_table': ( 'utils.html#modeltrackingcallback.log_metrics_table',
                                                                                              'fastMONAI/utils.py'),
                                 'fastMONAI.utils._detect_patch_workflow': ('utils.html#_detect_patch_workflow', 'fastMONAI/utils.py'),
                                 'fastMONAI.utils._extract_loss_name': ('utils.html#_extract_loss_name', 'fastMONAI/utils.py'),
                                 'fastMONAI.utils._extract_model_name': ('utils.html#_extract_model_name', 'fastMONAI/utils.py'),
                                 'fastMONAI.utils._extract_patch_config': ('utils.html#_extract_patch_config', 'fastMONAI/utils.py'),
                                 'fastMONAI.utils._extract_size_from_transforms': ( 'utils.html#_extract_size_from_transforms',
                                                                                    'fastMONAI/utils.py'),
                                 'fastMONAI.utils._extract_standard_config': ('utils.html#_extract_standard_config', 'fastMONAI/utils.py'),
                                 'fastMONAI.utils.create_mlflow_callback': ('utils.html#create_mlflow_callback', 'fastMONAI/utils.py'),
                                 'fastMONAI.utils.load_patch_variables': ('utils.html#load_patch_variables', 'fastMONAI/utils.py'),
                                 'fastMONAI.utils.load_variables': ('utils.html#load_variables', 'fastMONAI/utils.py'),
                                 'fastMONAI.utils.print_colab_gpu_info': ('utils.html#print_colab_gpu_info', 'fastMONAI/utils.py'),
                                 'fastMONAI.utils.store_patch_variables': ('utils.html#store_patch_variables', 'fastMONAI/utils.py'),
                                 'fastMONAI.utils.store_variables': ('utils.html#store_variables', 'fastMONAI/utils.py')},
            'fastMONAI.vision_all': {},
            'fastMONAI.vision_augmentation': { 'fastMONAI.vision_augmentation.BinaryConverter': ( 'vision_augment.html#binaryconverter',
                                                                                                  'fastMONAI/vision_augmentation.py'),
                                               'fastMONAI.vision_augmentation.BinaryConverter.encodes': ( 'vision_augment.html#binaryconverter.encodes',
                                                                                                          'fastMONAI/vision_augmentation.py'),
                                               'fastMONAI.vision_augmentation.BraTSMaskConverter': ( 'vision_augment.html#bratsmaskconverter',
                                                                                                     'fastMONAI/vision_augmentation.py'),
                                               'fastMONAI.vision_augmentation.BraTSMaskConverter.encodes': ( 'vision_augment.html#bratsmaskconverter.encodes',
                                                                                                             'fastMONAI/vision_augmentation.py'),
                                               'fastMONAI.vision_augmentation.CustomDictTransform': ( 'vision_augment.html#customdicttransform',
                                                                                                      'fastMONAI/vision_augmentation.py'),
                                               'fastMONAI.vision_augmentation.CustomDictTransform.__init__': ( 'vision_augment.html#customdicttransform.__init__',
                                                                                                               'fastMONAI/vision_augmentation.py'),
                                               'fastMONAI.vision_augmentation.CustomDictTransform.encodes': ( 'vision_augment.html#customdicttransform.encodes',
                                                                                                              'fastMONAI/vision_augmentation.py'),
                                               'fastMONAI.vision_augmentation.CustomDictTransform.tio_transform': ( 'vision_augment.html#customdicttransform.tio_transform',
                                                                                                                    'fastMONAI/vision_augmentation.py'),
                                               'fastMONAI.vision_augmentation.GpuPatchAugmentation': ( 'vision_augment.html#gpupatchaugmentation',
                                                                                                       'fastMONAI/vision_augmentation.py'),
                                               'fastMONAI.vision_augmentation.GpuPatchAugmentation.__call__': ( 'vision_augment.html#gpupatchaugmentation.__call__',
                                                                                                                'fastMONAI/vision_augmentation.py'),
                                               'fastMONAI.vision_augmentation.GpuPatchAugmentation.__init__': ( 'vision_augment.html#gpupatchaugmentation.__init__',
                                                                                                                'fastMONAI/vision_augmentation.py'),
                                               'fastMONAI.vision_augmentation.GpuPatchAugmentation.__repr__': ( 'vision_augment.html#gpupatchaugmentation.__repr__',
                                                                                                                'fastMONAI/vision_augmentation.py'),
                                               'fastMONAI.vision_augmentation.GpuPatchAugmentation._apply_affine': ( 'vision_augment.html#gpupatchaugmentation._apply_affine',
                                                                                                                     'fastMONAI/vision_augmentation.py'),
                                               'fastMONAI.vision_augmentation.GpuPatchAugmentation._apply_anisotropy': ( 'vision_augment.html#gpupatchaugmentation._apply_anisotropy',
                                                                                                                         'fastMONAI/vision_augmentation.py'),
                                               'fastMONAI.vision_augmentation.GpuPatchAugmentation._apply_blur': ( 'vision_augment.html#gpupatchaugmentation._apply_blur',
                                                                                                                   'fastMONAI/vision_augmentation.py'),
                                               'fastMONAI.vision_augmentation.GpuPatchAugmentation._apply_flip': ( 'vision_augment.html#gpupatchaugmentation._apply_flip',
                                                                                                                   'fastMONAI/vision_augmentation.py'),
                                               'fastMONAI.vision_augmentation.GpuPatchAugmentation._apply_gamma': ( 'vision_augment.html#gpupatchaugmentation._apply_gamma',
                                                                                                                    'fastMONAI/vision_augmentation.py'),
                                               'fastMONAI.vision_augmentation.GpuPatchAugmentation._apply_intensity_scale': ( 'vision_augment.html#gpupatchaugmentation._apply_intensity_scale',
                                                                                                                              'fastMONAI/vision_augmentation.py'),
                                               'fastMONAI.vision_augmentation.GpuPatchAugmentation._apply_noise': ( 'vision_augment.html#gpupatchaugmentation._apply_noise',
                                                                                                                    'fastMONAI/vision_augmentation.py'),
                                               'fastMONAI.vision_augmentation.NormalizeIntensity': ( 'vision_augment.html#normalizeintensity',
                                                                                                     'fastMONAI/vision_augmentation.py'),
                                               'fastMONAI.vision_augmentation.NormalizeIntensity.__init__': ( 'vision_augment.html#normalizeintensity.__init__',
                                                                                                              'fastMONAI/vision_augmentation.py'),
                                               'fastMONAI.vision_augmentation.NormalizeIntensity.encodes': ( 'vision_augment.html#normalizeintensity.encodes',
                                                                                                             'fastMONAI/vision_augmentation.py'),
                                               'fastMONAI.vision_augmentation.NormalizeIntensity.tio_transform': ( 'vision_augment.html#normalizeintensity.tio_transform',
                                                                                                                   'fastMONAI/vision_augmentation.py'),
                                               'fastMONAI.vision_augmentation.OneOf': ( 'vision_augment.html#oneof',
                                                                                        'fastMONAI/vision_augmentation.py'),
                                               'fastMONAI.vision_augmentation.OneOf.__init__': ( 'vision_augment.html#oneof.__init__',
                                                                                                 'fastMONAI/vision_augmentation.py'),
                                               'fastMONAI.vision_augmentation.PadOrCrop': ( 'vision_augment.html#padorcrop',
                                                                                            'fastMONAI/vision_augmentation.py'),
                                               'fastMONAI.vision_augmentation.PadOrCrop.__init__': ( 'vision_augment.html#padorcrop.__init__',
                                                                                                     'fastMONAI/vision_augmentation.py'),
                                               'fastMONAI.vision_augmentation.PadOrCrop.encodes': ( 'vision_augment.html#padorcrop.encodes',
                                                                                                    'fastMONAI/vision_augmentation.py'),
                                               'fastMONAI.vision_augmentation.PadOrCrop.tio_transform': ( 'vision_augment.html#padorcrop.tio_transform',
                                                                                                          'fastMONAI/vision_augmentation.py'),
                                               'fastMONAI.vision_augmentation.RandomAffine': ( 'vision_augment.html#randomaffine',
                                                                                               'fastMONAI/vision_augmentation.py'),
                                               'fastMONAI.vision_augmentation.RandomAffine.__init__': ( 'vision_augment.html#randomaffine.__init__',
                                                                                                        'fastMONAI/vision_augmentation.py'),
                                               'fastMONAI.vision_augmentation.RandomAnisotropy': ( 'vision_augment.html#randomanisotropy',
                                                                                                   'fastMONAI/vision_augmentation.py'),
                                               'fastMONAI.vision_augmentation.RandomAnisotropy.__init__': ( 'vision_augment.html#randomanisotropy.__init__',
                                                                                                            'fastMONAI/vision_augmentation.py'),
                                               'fastMONAI.vision_augmentation.RandomAnisotropy.encodes': ( 'vision_augment.html#randomanisotropy.encodes',
                                                                                                           'fastMONAI/vision_augmentation.py'),
                                               'fastMONAI.vision_augmentation.RandomAnisotropy.tio_transform': ( 'vision_augment.html#randomanisotropy.tio_transform',
                                                                                                                 'fastMONAI/vision_augmentation.py'),
                                               'fastMONAI.vision_augmentation.RandomBiasField': ( 'vision_augment.html#randombiasfield',
                                                                                                  'fastMONAI/vision_augmentation.py'),
                                               'fastMONAI.vision_augmentation.RandomBiasField.__init__': ( 'vision_augment.html#randombiasfield.__init__',
                                                                                                           'fastMONAI/vision_augmentation.py'),
                                               'fastMONAI.vision_augmentation.RandomBiasField.encodes': ( 'vision_augment.html#randombiasfield.encodes',
                                                                                                          'fastMONAI/vision_augmentation.py'),
                                               'fastMONAI.vision_augmentation.RandomBiasField.tio_transform': ( 'vision_augment.html#randombiasfield.tio_transform',
                                                                                                                'fastMONAI/vision_augmentation.py'),
                                               'fastMONAI.vision_augmentation.RandomBlur': ( 'vision_augment.html#randomblur',
                                                                                             'fastMONAI/vision_augmentation.py'),
                                               'fastMONAI.vision_augmentation.RandomBlur.__init__': ( 'vision_augment.html#randomblur.__init__',
                                                                                                      'fastMONAI/vision_augmentation.py'),
                                               'fastMONAI.vision_augmentation.RandomBlur.encodes': ( 'vision_augment.html#randomblur.encodes',
                                                                                                     'fastMONAI/vision_augmentation.py'),
                                               'fastMONAI.vision_augmentation.RandomBlur.tio_transform': ( 'vision_augment.html#randomblur.tio_transform',
                                                                                                           'fastMONAI/vision_augmentation.py'),
                                               'fastMONAI.vision_augmentation.RandomCutout': ( 'vision_augment.html#randomcutout',
                                                                                               'fastMONAI/vision_augmentation.py'),
                                               'fastMONAI.vision_augmentation.RandomCutout.__init__': ( 'vision_augment.html#randomcutout.__init__',
                                                                                                        'fastMONAI/vision_augmentation.py'),
                                               'fastMONAI.vision_augmentation.RandomCutout._get_fill_value': ( 'vision_augment.html#randomcutout._get_fill_value',
                                                                                                               'fastMONAI/vision_augmentation.py'),
                                               'fastMONAI.vision_augmentation.RandomCutout.encodes': ( 'vision_augment.html#randomcutout.encodes',
                                                                                                       'fastMONAI/vision_augmentation.py'),
                                               'fastMONAI.vision_augmentation.RandomCutout.tio_transform': ( 'vision_augment.html#randomcutout.tio_transform',
                                                                                                             'fastMONAI/vision_augmentation.py'),
                                               'fastMONAI.vision_augmentation.RandomElasticDeformation': ( 'vision_augment.html#randomelasticdeformation',
                                                                                                           'fastMONAI/vision_augmentation.py'),
                                               'fastMONAI.vision_augmentation.RandomElasticDeformation.__init__': ( 'vision_augment.html#randomelasticdeformation.__init__',
                                                                                                                    'fastMONAI/vision_augmentation.py'),
                                               'fastMONAI.vision_augmentation.RandomFlip': ( 'vision_augment.html#randomflip',
                                                                                             'fastMONAI/vision_augmentation.py'),
                                               'fastMONAI.vision_augmentation.RandomFlip.__init__': ( 'vision_augment.html#randomflip.__init__',
                                                                                                      'fastMONAI/vision_augmentation.py'),
                                               'fastMONAI.vision_augmentation.RandomGamma': ( 'vision_augment.html#randomgamma',
                                                                                              'fastMONAI/vision_augmentation.py'),
                                               'fastMONAI.vision_augmentation.RandomGamma.__init__': ( 'vision_augment.html#randomgamma.__init__',
                                                                                                       'fastMONAI/vision_augmentation.py'),
                                               'fastMONAI.vision_augmentation.RandomGamma.encodes': ( 'vision_augment.html#randomgamma.encodes',
                                                                                                      'fastMONAI/vision_augmentation.py'),
                                               'fastMONAI.vision_augmentation.RandomGamma.tio_transform': ( 'vision_augment.html#randomgamma.tio_transform',
                                                                                                            'fastMONAI/vision_augmentation.py'),
                                               'fastMONAI.vision_augmentation.RandomGhosting': ( 'vision_augment.html#randomghosting',
                                                                                                 'fastMONAI/vision_augmentation.py'),
                                               'fastMONAI.vision_augmentation.RandomGhosting.__init__': ( 'vision_augment.html#randomghosting.__init__',
                                                                                                          'fastMONAI/vision_augmentation.py'),
                                               'fastMONAI.vision_augmentation.RandomGhosting.encodes': ( 'vision_augment.html#randomghosting.encodes',
                                                                                                         'fastMONAI/vision_augmentation.py'),
                                               'fastMONAI.vision_augmentation.RandomGhosting.tio_transform': ( 'vision_augment.html#randomghosting.tio_transform',
                                                                                                               'fastMONAI/vision_augmentation.py'),
                                               'fastMONAI.vision_augmentation.RandomIntensityScale': ( 'vision_augment.html#randomintensityscale',
                                                                                                       'fastMONAI/vision_augmentation.py'),
                                               'fastMONAI.vision_augmentation.RandomIntensityScale.__init__': ( 'vision_augment.html#randomintensityscale.__init__',
                                                                                                                'fastMONAI/vision_augmentation.py'),
                                               'fastMONAI.vision_augmentation.RandomIntensityScale.encodes': ( 'vision_augment.html#randomintensityscale.encodes',
                                                                                                               'fastMONAI/vision_augmentation.py'),
                                               'fastMONAI.vision_augmentation.RandomIntensityScale.tio_transform': ( 'vision_augment.html#randomintensityscale.tio_transform',
                                                                                                                     'fastMONAI/vision_augmentation.py'),
                                               'fastMONAI.vision_augmentation.RandomMotion': ( 'vision_augment.html#randommotion',
                                                                                               'fastMONAI/vision_augmentation.py'),
                                               'fastMONAI.vision_augmentation.RandomMotion.__init__': ( 'vision_augment.html#randommotion.__init__',
                                                                                                        'fastMONAI/vision_augmentation.py'),
                                               'fastMONAI.vision_augmentation.RandomMotion.encodes': ( 'vision_augment.html#randommotion.encodes',
                                                                                                       'fastMONAI/vision_augmentation.py'),
                                               'fastMONAI.vision_augmentation.RandomMotion.tio_transform': ( 'vision_augment.html#randommotion.tio_transform',
                                                                                                             'fastMONAI/vision_augmentation.py'),
                                               'fastMONAI.vision_augmentation.RandomNoise': ( 'vision_augment.html#randomnoise',
                                                                                              'fastMONAI/vision_augmentation.py'),
                                               'fastMONAI.vision_augmentation.RandomNoise.__init__': ( 'vision_augment.html#randomnoise.__init__',
                                                                                                       'fastMONAI/vision_augmentation.py'),
                                               'fastMONAI.vision_augmentation.RandomNoise.encodes': ( 'vision_augment.html#randomnoise.encodes',
                                                                                                      'fastMONAI/vision_augmentation.py'),
                                               'fastMONAI.vision_augmentation.RandomNoise.tio_transform': ( 'vision_augment.html#randomnoise.tio_transform',
                                                                                                            'fastMONAI/vision_augmentation.py'),
                                               'fastMONAI.vision_augmentation.RandomSpike': ( 'vision_augment.html#randomspike',
                                                                                              'fastMONAI/vision_augmentation.py'),
                                               'fastMONAI.vision_augmentation.RandomSpike.__init__': ( 'vision_augment.html#randomspike.__init__',
                                                                                                       'fastMONAI/vision_augmentation.py'),
                                               'fastMONAI.vision_augmentation.RandomSpike.encodes': ( 'vision_augment.html#randomspike.encodes',
                                                                                                      'fastMONAI/vision_augmentation.py'),
                                               'fastMONAI.vision_augmentation.RandomSpike.tio_transform': ( 'vision_augment.html#randomspike.tio_transform',
                                                                                                            'fastMONAI/vision_augmentation.py'),
                                               'fastMONAI.vision_augmentation.RescaleIntensity': ( 'vision_augment.html#rescaleintensity',
                                                                                                   'fastMONAI/vision_augmentation.py'),
                                               'fastMONAI.vision_augmentation.RescaleIntensity.__init__': ( 'vision_augment.html#rescaleintensity.__init__',
                                                                                                            'fastMONAI/vision_augmentation.py'),
                                               'fastMONAI.vision_augmentation.RescaleIntensity.encodes': ( 'vision_augment.html#rescaleintensity.encodes',
                                                                                                           'fastMONAI/vision_augmentation.py'),
                                               'fastMONAI.vision_augmentation.RescaleIntensity.tio_transform': ( 'vision_augment.html#rescaleintensity.tio_transform',
                                                                                                                 'fastMONAI/vision_augmentation.py'),
                                               'fastMONAI.vision_augmentation.ZNormalization': ( 'vision_augment.html#znormalization',
                                                                                                 'fastMONAI/vision_augmentation.py'),
                                               'fastMONAI.vision_augmentation.ZNormalization.__init__': ( 'vision_augment.html#znormalization.__init__',
                                                                                                          'fastMONAI/vision_augmentation.py'),
                                               'fastMONAI.vision_augmentation.ZNormalization.encodes': ( 'vision_augment.html#znormalization.encodes',
                                                                                                         'fastMONAI/vision_augmentation.py'),
                                               'fastMONAI.vision_augmentation.ZNormalization.tio_transform': ( 'vision_augment.html#znormalization.tio_transform',
                                                                                                               'fastMONAI/vision_augmentation.py'),
                                               'fastMONAI.vision_augmentation._TioNormalizeIntensity': ( 'vision_augment.html#_tionormalizeintensity',
                                                                                                         'fastMONAI/vision_augmentation.py'),
                                               'fastMONAI.vision_augmentation._TioNormalizeIntensity.__init__': ( 'vision_augment.html#_tionormalizeintensity.__init__',
                                                                                                                  'fastMONAI/vision_augmentation.py'),
                                               'fastMONAI.vision_augmentation._TioNormalizeIntensity.apply_transform': ( 'vision_augment.html#_tionormalizeintensity.apply_transform',
                                                                                                                         'fastMONAI/vision_augmentation.py'),
                                               'fastMONAI.vision_augmentation._TioRandomCutout': ( 'vision_augment.html#_tiorandomcutout',
                                                                                                   'fastMONAI/vision_augmentation.py'),
                                               'fastMONAI.vision_augmentation._TioRandomCutout.__init__': ( 'vision_augment.html#_tiorandomcutout.__init__',
                                                                                                            'fastMONAI/vision_augmentation.py'),
                                               'fastMONAI.vision_augmentation._TioRandomCutout._apply_cutout': ( 'vision_augment.html#_tiorandomcutout._apply_cutout',
                                                                                                                 'fastMONAI/vision_augmentation.py'),
                                               'fastMONAI.vision_augmentation._TioRandomCutout.apply_transform': ( 'vision_augment.html#_tiorandomcutout.apply_transform',
                                                                                                                   'fastMONAI/vision_augmentation.py'),
                                               'fastMONAI.vision_augmentation._TioRandomIntensityScale': ( 'vision_augment.html#_tiorandomintensityscale',
                                                                                                           'fastMONAI/vision_augmentation.py'),
                                               'fastMONAI.vision_augmentation._TioRandomIntensityScale.__init__': ( 'vision_augment.html#_tiorandomintensityscale.__init__',
                                                                                                                    'fastMONAI/vision_augmentation.py'),
                                               'fastMONAI.vision_augmentation._TioRandomIntensityScale.apply_transform': ( 'vision_augment.html#_tiorandomintensityscale.apply_transform',
                                                                                                                           'fastMONAI/vision_augmentation.py'),
                                               'fastMONAI.vision_augmentation._build_rotation_matrix_3d': ( 'vision_augment.html#_build_rotation_matrix_3d',
                                                                                                            'fastMONAI/vision_augmentation.py'),
                                               'fastMONAI.vision_augmentation._compute_patch_aug_params': ( 'vision_augment.html#_compute_patch_aug_params',
                                                                                                            'fastMONAI/vision_augmentation.py'),
                                               'fastMONAI.vision_augmentation._create_ellipsoid_mask': ( 'vision_augment.html#_create_ellipsoid_mask',
                                                                                                         'fastMONAI/vision_augmentation.py'),
                                               'fastMONAI.vision_augmentation._foreground_masking': ( 'vision_augment.html#_foreground_masking',
                                                                                                      'fastMONAI/vision_augmentation.py'),
                                               'fastMONAI.vision_augmentation.do_pad_or_crop': ( 'vision_augment.html#do_pad_or_crop',
                                                                                                 'fastMONAI/vision_augmentation.py'),
                                               'fastMONAI.vision_augmentation.gpu_patch_augmentations': ( 'vision_augment.html#gpu_patch_augmentations',
                                                                                                          'fastMONAI/vision_augmentation.py'),
                                               'fastMONAI.vision_augmentation.suggest_patch_augmentations': ( 'vision_augment.html#suggest_patch_augmentations',
                                                                                                              'fastMONAI/vision_augmentation.py')},
            'fastMONAI.vision_core': { 'fastMONAI.vision_core.MedBase': ('vision_core.html#medbase', 'fastMONAI/vision_core.py'),
                                       'fastMONAI.vision_core.MedBase.__copy__': ( 'vision_core.html#medbase.__copy__',
                                                                                   'fastMONAI/vision_core.py'),
                                       'fastMONAI.vision_core.MedBase.__deepcopy__': ( 'vision_core.html#medbase.__deepcopy__',
                                                                                       'fastMONAI/vision_core.py'),
                                       'fastMONAI.vision_core.MedBase.__new__': ( 'vision_core.html#medbase.__new__',
                                                                                  'fastMONAI/vision_core.py'),
                                       'fastMONAI.vision_core.MedBase.__repr__': ( 'vision_core.html#medbase.__repr__',
                                                                                   'fastMONAI/vision_core.py'),
                                       'fastMONAI.vision_core.MedBase.create': ( 'vision_core.html#medbase.create',
                                                                                 'fastMONAI/vision_core.py'),
                                       'fastMONAI.vision_core.MedBase.item_preprocessing': ( 'vision_core.html#medbase.item_preprocessing',
                                                                                             'fastMONAI/vision_core.py'),
                                       'fastMONAI.vision_core.MedBase.new_empty': ( 'vision_core.html#medbase.new_empty',
                                                                                    'fastMONAI/vision_core.py'),
                                       'fastMONAI.vision_core.MedBase.show': ('vision_core.html#medbase.show', 'fastMONAI/vision_core.py'),
                                       'fastMONAI.vision_core.MedImage': ('vision_core.html#medimage', 'fastMONAI/vision_core.py'),
                                       'fastMONAI.vision_core.MedMask': ('vision_core.html#medmask', 'fastMONAI/vision_core.py'),
                                       'fastMONAI.vision_core.MetaResolver': ('vision_core.html#metaresolver', 'fastMONAI/vision_core.py'),
                                       'fastMONAI.vision_core._load_and_preprocess': ( 'vision_core.html#_load_and_preprocess',
                                                                                       'fastMONAI/vision_core.py'),
                                       'fastMONAI.vision_core._multi_channel': ( 'vision_core.html#_multi_channel',
                                                                                 'fastMONAI/vision_core.py'),
                                       'fastMONAI.vision_core._preprocess': ('vision_core.html#_preprocess', 'fastMONAI/vision_core.py'),
                                       'fastMONAI.vision_core.med_img_reader': ( 'vision_core.html#med_img_reader',
                                                                                 'fastMONAI/vision_core.py')},
            'fastMONAI.vision_data': { 'fastMONAI.vision_data.MedDataBlock': ('vision_data.html#meddatablock', 'fastMONAI/vision_data.py'),
                                       'fastMONAI.vision_data.MedDataBlock.__init__': ( 'vision_data.html#meddatablock.__init__',
                                                                                        'fastMONAI/vision_data.py'),
                                       'fastMONAI.vision_data.MedImageDataLoaders': ( 'vision_data.html#medimagedataloaders',
                                                                                      'fastMONAI/vision_data.py'),
                                       'fastMONAI.vision_data.MedImageDataLoaders.from_df': ( 'vision_data.html#medimagedataloaders.from_df',
                                                                                              'fastMONAI/vision_data.py'),
                                       'fastMONAI.vision_data.MedMaskBlock': ('vision_data.html#medmaskblock', 'fastMONAI/vision_data.py'),
                                       'fastMONAI.vision_data.batch_pred_to_multiclass_mask': ( 'vision_data.html#batch_pred_to_multiclass_mask',
                                                                                                'fastMONAI/vision_data.py'),
                                       'fastMONAI.vision_data.plot_top_losses': ( 'vision_data.html#plot_top_losses',
                                                                                  'fastMONAI/vision_data.py'),
                                       'fastMONAI.vision_data.pred_to_binary_mask': ( 'vision_data.html#pred_to_binary_mask',
                                                                                      'fastMONAI/vision_data.py'),
                                       'fastMONAI.vision_data.pred_to_multiclass_mask': ( 'vision_data.html#pred_to_multiclass_mask',
                                                                                          'fastMONAI/vision_data.py'),
                                       'fastMONAI.vision_data.show_batch': ('vision_data.html#show_batch', 'fastMONAI/vision_data.py'),
                                       'fastMONAI.vision_data.show_results': ('vision_data.html#show_results', 'fastMONAI/vision_data.py')},
            'fastMONAI.vision_inference': { 'fastMONAI.vision_inference._do_resize': ( 'vision_inference.html#_do_resize',
                                                                                       'fastMONAI/vision_inference.py'),
                                            'fastMONAI.vision_inference._to_original_orientation': ( 'vision_inference.html#_to_original_orientation',
                                                                                                     'fastMONAI/vision_inference.py'),
                                            'fastMONAI.vision_inference._update_uid': ( 'vision_inference.html#_update_uid',
                                                                                        'fastMONAI/vision_inference.py'),
                                            'fastMONAI.vision_inference.compute_binary_tumor_volume': ( 'vision_inference.html#compute_binary_tumor_volume',
                                                                                                        'fastMONAI/vision_inference.py'),
                                            'fastMONAI.vision_inference.gradio_image_classifier': ( 'vision_inference.html#gradio_image_classifier',
                                                                                                    'fastMONAI/vision_inference.py'),
                                            'fastMONAI.vision_inference.inference': ( 'vision_inference.html#inference',
                                                                                      'fastMONAI/vision_inference.py'),
                                            'fastMONAI.vision_inference.keep_largest': ( 'vision_inference.html#keep_largest',
                                                                                         'fastMONAI/vision_inference.py'),
                                            'fastMONAI.vision_inference.load_system_resources': ( 'vision_inference.html#load_system_resources',
                                                                                                  'fastMONAI/vision_inference.py'),
                                            'fastMONAI.vision_inference.refine_binary_pred_mask': ( 'vision_inference.html#refine_binary_pred_mask',
                                                                                                    'fastMONAI/vision_inference.py'),
                                            'fastMONAI.vision_inference.save_series_pred': ( 'vision_inference.html#save_series_pred',
                                                                                             'fastMONAI/vision_inference.py')},
            'fastMONAI.vision_loss': { 'fastMONAI.vision_loss.CustomLoss': ( 'vision_loss_functions.html#customloss',
                                                                             'fastMONAI/vision_loss.py'),
                                       'fastMONAI.vision_loss.CustomLoss.__call__': ( 'vision_loss_functions.html#customloss.__call__',
                                                                                      'fastMONAI/vision_loss.py'),
                                       'fastMONAI.vision_loss.CustomLoss.__init__': ( 'vision_loss_functions.html#customloss.__init__',
                                                                                      'fastMONAI/vision_loss.py'),
                                       'fastMONAI.vision_loss.CustomLoss.activation': ( 'vision_loss_functions.html#customloss.activation',
                                                                                        'fastMONAI/vision_loss.py'),
                                       'fastMONAI.vision_loss.CustomLoss.decodes': ( 'vision_loss_functions.html#customloss.decodes',
                                                                                     'fastMONAI/vision_loss.py'),
                                       'fastMONAI.vision_loss.TverskyFocalLoss': ( 'vision_loss_functions.html#tverskyfocalloss',
                                                                                   'fastMONAI/vision_loss.py'),
                                       'fastMONAI.vision_loss.TverskyFocalLoss.__init__': ( 'vision_loss_functions.html#tverskyfocalloss.__init__',
                                                                                            'fastMONAI/vision_loss.py'),
                                       'fastMONAI.vision_loss.TverskyFocalLoss.forward': ( 'vision_loss_functions.html#tverskyfocalloss.forward',
                                                                                           'fastMONAI/vision_loss.py')},
            'fastMONAI.vision_metrics': { 'fastMONAI.vision_metrics.AccumulatedDice': ( 'vision_metrics.html#accumulateddice',
                                                                                        'fastMONAI/vision_metrics.py'),
                                          'fastMONAI.vision_metrics.AccumulatedDice.__init__': ( 'vision_metrics.html#accumulateddice.__init__',
                                                                                                 'fastMONAI/vision_metrics.py'),
                                          'fastMONAI.vision_metrics.AccumulatedDice.accumulate': ( 'vision_metrics.html#accumulateddice.accumulate',
                                                                                                   'fastMONAI/vision_metrics.py'),
                                          'fastMONAI.vision_metrics.AccumulatedDice.name': ( 'vision_metrics.html#accumulateddice.name',
                                                                                             'fastMONAI/vision_metrics.py'),
                                          'fastMONAI.vision_metrics.AccumulatedDice.reset': ( 'vision_metrics.html#accumulateddice.reset',
                                                                                              'fastMONAI/vision_metrics.py'),
                                          'fastMONAI.vision_metrics.AccumulatedDice.value': ( 'vision_metrics.html#accumulateddice.value',
                                                                                              'fastMONAI/vision_metrics.py'),
                                          'fastMONAI.vision_metrics.AccumulatedMultiDice': ( 'vision_metrics.html#accumulatedmultidice',
                                                                                             'fastMONAI/vision_metrics.py'),
                                          'fastMONAI.vision_metrics.AccumulatedMultiDice.name': ( 'vision_metrics.html#accumulatedmultidice.name',
                                                                                                  'fastMONAI/vision_metrics.py'),
                                          'fastMONAI.vision_metrics.AccumulatedMultiDice.value': ( 'vision_metrics.html#accumulatedmultidice.value',
                                                                                                   'fastMONAI/vision_metrics.py'),
                                          'fastMONAI.vision_metrics.EMACheckpoint': ( 'vision_metrics.html#emacheckpoint',
                                                                                      'fastMONAI/vision_metrics.py'),
                                          'fastMONAI.vision_metrics.EMACheckpoint.__init__': ( 'vision_metrics.html#emacheckpoint.__init__',
                                                                                               'fastMONAI/vision_metrics.py'),
                                          'fastMONAI.vision_metrics.EMACheckpoint.after_epoch': ( 'vision_metrics.html#emacheckpoint.after_epoch',
                                                                                                  'fastMONAI/vision_metrics.py'),
                                          'fastMONAI.vision_metrics.EMACheckpoint.before_fit': ( 'vision_metrics.html#emacheckpoint.before_fit',
                                                                                                 'fastMONAI/vision_metrics.py'),
                                          'fastMONAI.vision_metrics.binary_dice_score': ( 'vision_metrics.html#binary_dice_score',
                                                                                          'fastMONAI/vision_metrics.py'),
                                          'fastMONAI.vision_metrics.binary_hausdorff_distance': ( 'vision_metrics.html#binary_hausdorff_distance',
                                                                                                  'fastMONAI/vision_metrics.py'),
                                          'fastMONAI.vision_metrics.binary_lesion_detection_rate': ( 'vision_metrics.html#binary_lesion_detection_rate',
                                                                                                     'fastMONAI/vision_metrics.py'),
                                          'fastMONAI.vision_metrics.binary_precision': ( 'vision_metrics.html#binary_precision',
                                                                                         'fastMONAI/vision_metrics.py'),
                                          'fastMONAI.vision_metrics.binary_sensitivity': ( 'vision_metrics.html#binary_sensitivity',
                                                                                           'fastMONAI/vision_metrics.py'),
                                          'fastMONAI.vision_metrics.binary_signed_rve': ( 'vision_metrics.html#binary_signed_rve',
                                                                                          'fastMONAI/vision_metrics.py'),
                                          'fastMONAI.vision_metrics.calculate_confusion_metrics': ( 'vision_metrics.html#calculate_confusion_metrics',
                                                                                                    'fastMONAI/vision_metrics.py'),
                                          'fastMONAI.vision_metrics.calculate_dsc': ( 'vision_metrics.html#calculate_dsc',
                                                                                      'fastMONAI/vision_metrics.py'),
                                          'fastMONAI.vision_metrics.calculate_haus': ( 'vision_metrics.html#calculate_haus',
                                                                                       'fastMONAI/vision_metrics.py'),
                                          'fastMONAI.vision_metrics.calculate_lesion_detection_rate': ( 'vision_metrics.html#calculate_lesion_detection_rate',
                                                                                                        'fastMONAI/vision_metrics.py'),
                                          'fastMONAI.vision_metrics.calculate_signed_rve': ( 'vision_metrics.html#calculate_signed_rve',
                                                                                             'fastMONAI/vision_metrics.py'),
                                          'fastMONAI.vision_metrics.multi_dice_score': ( 'vision_metrics.html#multi_dice_score',
                                                                                         'fastMONAI/vision_metrics.py'),
                                          'fastMONAI.vision_metrics.multi_hausdorff_distance': ( 'vision_metrics.html#multi_hausdorff_distance',
                                                                                                 'fastMONAI/vision_metrics.py'),
                                          'fastMONAI.vision_metrics.multi_lesion_detection_rate': ( 'vision_metrics.html#multi_lesion_detection_rate',
                                                                                                    'fastMONAI/vision_metrics.py'),
                                          'fastMONAI.vision_metrics.multi_precision': ( 'vision_metrics.html#multi_precision',
                                                                                        'fastMONAI/vision_metrics.py'),
                                          'fastMONAI.vision_metrics.multi_sensitivity': ( 'vision_metrics.html#multi_sensitivity',
                                                                                          'fastMONAI/vision_metrics.py'),
                                          'fastMONAI.vision_metrics.multi_signed_rve': ( 'vision_metrics.html#multi_signed_rve',
                                                                                         'fastMONAI/vision_metrics.py')},
            'fastMONAI.vision_patch': { 'fastMONAI.vision_patch.MedPatchDataLoader': ( 'vision_patch.html#medpatchdataloader',
                                                                                       'fastMONAI/vision_patch.py'),
                                        'fastMONAI.vision_patch.MedPatchDataLoader.__del__': ( 'vision_patch.html#medpatchdataloader.__del__',
                                                                                               'fastMONAI/vision_patch.py'),
                                        'fastMONAI.vision_patch.MedPatchDataLoader.__enter__': ( 'vision_patch.html#medpatchdataloader.__enter__',
                                                                                                 'fastMONAI/vision_patch.py'),
                                        'fastMONAI.vision_patch.MedPatchDataLoader.__exit__': ( 'vision_patch.html#medpatchdataloader.__exit__',
                                                                                                'fastMONAI/vision_patch.py'),
                                        'fastMONAI.vision_patch.MedPatchDataLoader.__init__': ( 'vision_patch.html#medpatchdataloader.__init__',
                                                                                                'fastMONAI/vision_patch.py'),
                                        'fastMONAI.vision_patch.MedPatchDataLoader.__iter__': ( 'vision_patch.html#medpatchdataloader.__iter__',
                                                                                                'fastMONAI/vision_patch.py'),
                                        'fastMONAI.vision_patch.MedPatchDataLoader.__len__': ( 'vision_patch.html#medpatchdataloader.__len__',
                                                                                               'fastMONAI/vision_patch.py'),
                                        'fastMONAI.vision_patch.MedPatchDataLoader.close': ( 'vision_patch.html#medpatchdataloader.close',
                                                                                             'fastMONAI/vision_patch.py'),
                                        'fastMONAI.vision_patch.MedPatchDataLoader.dataset': ( 'vision_patch.html#medpatchdataloader.dataset',
                                                                                               'fastMONAI/vision_patch.py'),
                                        'fastMONAI.vision_patch.MedPatchDataLoader.device': ( 'vision_patch.html#medpatchdataloader.device',
                                                                                              'fastMONAI/vision_patch.py'),
                                        'fastMONAI.vision_patch.MedPatchDataLoader.one_batch': ( 'vision_patch.html#medpatchdataloader.one_batch',
                                                                                                 'fastMONAI/vision_patch.py'),
                                        'fastMONAI.vision_patch.MedPatchDataLoader.to': ( 'vision_patch.html#medpatchdataloader.to',
                                                                                          'fastMONAI/vision_patch.py'),
                                        'fastMONAI.vision_patch.MedPatchDataLoaders': ( 'vision_patch.html#medpatchdataloaders',
                                                                                        'fastMONAI/vision_patch.py'),
                                        'fastMONAI.vision_patch.MedPatchDataLoaders.__del__': ( 'vision_patch.html#medpatchdataloaders.__del__',
                                                                                                'fastMONAI/vision_patch.py'),
                                        'fastMONAI.vision_patch.MedPatchDataLoaders.__enter__': ( 'vision_patch.html#medpatchdataloaders.__enter__',
                                                                                                  'fastMONAI/vision_patch.py'),
                                        'fastMONAI.vision_patch.MedPatchDataLoaders.__exit__': ( 'vision_patch.html#medpatchdataloaders.__exit__',
                                                                                                 'fastMONAI/vision_patch.py'),
                                        'fastMONAI.vision_patch.MedPatchDataLoaders.__getitem__': ( 'vision_patch.html#medpatchdataloaders.__getitem__',
                                                                                                    'fastMONAI/vision_patch.py'),
                                        'fastMONAI.vision_patch.MedPatchDataLoaders.__init__': ( 'vision_patch.html#medpatchdataloaders.__init__',
                                                                                                 'fastMONAI/vision_patch.py'),
                                        'fastMONAI.vision_patch.MedPatchDataLoaders.__iter__': ( 'vision_patch.html#medpatchdataloaders.__iter__',
                                                                                                 'fastMONAI/vision_patch.py'),
                                        'fastMONAI.vision_patch.MedPatchDataLoaders.__len__': ( 'vision_patch.html#medpatchdataloaders.__len__',
                                                                                                'fastMONAI/vision_patch.py'),
                                        'fastMONAI.vision_patch.MedPatchDataLoaders.apply_reorder': ( 'vision_patch.html#medpatchdataloaders.apply_reorder',
                                                                                                      'fastMONAI/vision_patch.py'),
                                        'fastMONAI.vision_patch.MedPatchDataLoaders.bs': ( 'vision_patch.html#medpatchdataloaders.bs',
                                                                                           'fastMONAI/vision_patch.py'),
                                        'fastMONAI.vision_patch.MedPatchDataLoaders.close': ( 'vision_patch.html#medpatchdataloaders.close',
                                                                                              'fastMONAI/vision_patch.py'),
                                        'fastMONAI.vision_patch.MedPatchDataLoaders.cpu': ( 'vision_patch.html#medpatchdataloaders.cpu',
                                                                                            'fastMONAI/vision_patch.py'),
                                        'fastMONAI.vision_patch.MedPatchDataLoaders.cuda': ( 'vision_patch.html#medpatchdataloaders.cuda',
                                                                                             'fastMONAI/vision_patch.py'),
                                        'fastMONAI.vision_patch.MedPatchDataLoaders.device': ( 'vision_patch.html#medpatchdataloaders.device',
                                                                                               'fastMONAI/vision_patch.py'),
                                        'fastMONAI.vision_patch.MedPatchDataLoaders.from_df': ( 'vision_patch.html#medpatchdataloaders.from_df',
                                                                                                'fastMONAI/vision_patch.py'),
                                        'fastMONAI.vision_patch.MedPatchDataLoaders.new_empty': ( 'vision_patch.html#medpatchdataloaders.new_empty',
                                                                                                  'fastMONAI/vision_patch.py'),
                                        'fastMONAI.vision_patch.MedPatchDataLoaders.one_batch': ( 'vision_patch.html#medpatchdataloaders.one_batch',
                                                                                                  'fastMONAI/vision_patch.py'),
                                        'fastMONAI.vision_patch.MedPatchDataLoaders.patch_config': ( 'vision_patch.html#medpatchdataloaders.patch_config',
                                                                                                     'fastMONAI/vision_patch.py'),
                                        'fastMONAI.vision_patch.MedPatchDataLoaders.show_batch': ( 'vision_patch.html#medpatchdataloaders.show_batch',
                                                                                                   'fastMONAI/vision_patch.py'),
                                        'fastMONAI.vision_patch.MedPatchDataLoaders.split_df': ( 'vision_patch.html#medpatchdataloaders.split_df',
                                                                                                 'fastMONAI/vision_patch.py'),
                                        'fastMONAI.vision_patch.MedPatchDataLoaders.target_spacing': ( 'vision_patch.html#medpatchdataloaders.target_spacing',
                                                                                                       'fastMONAI/vision_patch.py'),
                                        'fastMONAI.vision_patch.MedPatchDataLoaders.to': ( 'vision_patch.html#medpatchdataloaders.to',
                                                                                           'fastMONAI/vision_patch.py'),
                                        'fastMONAI.vision_patch.MedPatchDataLoaders.train': ( 'vision_patch.html#medpatchdataloaders.train',
                                                                                              'fastMONAI/vision_patch.py'),
                                        'fastMONAI.vision_patch.MedPatchDataLoaders.train_ds': ( 'vision_patch.html#medpatchdataloaders.train_ds',
                                                                                                 'fastMONAI/vision_patch.py'),
                                        'fastMONAI.vision_patch.MedPatchDataLoaders.valid': ( 'vision_patch.html#medpatchdataloaders.valid',
                                                                                              'fastMONAI/vision_patch.py'),
                                        'fastMONAI.vision_patch.MedPatchDataLoaders.valid_ds': ( 'vision_patch.html#medpatchdataloaders.valid_ds',
                                                                                                 'fastMONAI/vision_patch.py'),
                                        'fastMONAI.vision_patch.PatchConfig': ( 'vision_patch.html#patchconfig',
                                                                                'fastMONAI/vision_patch.py'),
                                        'fastMONAI.vision_patch.PatchConfig.__post_init__': ( 'vision_patch.html#patchconfig.__post_init__',
                                                                                              'fastMONAI/vision_patch.py'),
                                        'fastMONAI.vision_patch.PatchConfig.from_dataset': ( 'vision_patch.html#patchconfig.from_dataset',
                                                                                             'fastMONAI/vision_patch.py'),
                                        'fastMONAI.vision_patch.PatchInferenceEngine': ( 'vision_patch.html#patchinferenceengine',
                                                                                         'fastMONAI/vision_patch.py'),
                                        'fastMONAI.vision_patch.PatchInferenceEngine.__init__': ( 'vision_patch.html#patchinferenceengine.__init__',
                                                                                                  'fastMONAI/vision_patch.py'),
                                        'fastMONAI.vision_patch.PatchInferenceEngine.predict': ( 'vision_patch.html#patchinferenceengine.predict',
                                                                                                 'fastMONAI/vision_patch.py'),
                                        'fastMONAI.vision_patch.PatchInferenceEngine.to': ( 'vision_patch.html#patchinferenceengine.to',
                                                                                            'fastMONAI/vision_patch.py'),
                                        'fastMONAI.vision_patch._extract_tio_transform': ( 'vision_patch.html#_extract_tio_transform',
                                                                                           'fastMONAI/vision_patch.py'),
                                        'fastMONAI.vision_patch._get_default_device': ( 'vision_patch.html#_get_default_device',
                                                                                        'fastMONAI/vision_patch.py'),
                                        'fastMONAI.vision_patch._normalize_patch_overlap': ( 'vision_patch.html#_normalize_patch_overlap',
                                                                                             'fastMONAI/vision_patch.py'),
                                        'fastMONAI.vision_patch._predict_patch_tta': ( 'vision_patch.html#_predict_patch_tta',
                                                                                       'fastMONAI/vision_patch.py'),
                                        'fastMONAI.vision_patch._warn_config_override': ( 'vision_patch.html#_warn_config_override',
                                                                                          'fastMONAI/vision_patch.py'),
                                        'fastMONAI.vision_patch.create_patch_sampler': ( 'vision_patch.html#create_patch_sampler',
                                                                                         'fastMONAI/vision_patch.py'),
                                        'fastMONAI.vision_patch.create_subjects_dataset': ( 'vision_patch.html#create_subjects_dataset',
                                                                                            'fastMONAI/vision_patch.py'),
                                        'fastMONAI.vision_patch.med_to_subject': ( 'vision_patch.html#med_to_subject',
                                                                                   'fastMONAI/vision_patch.py'),
                                        'fastMONAI.vision_patch.normalize_patch_transforms': ( 'vision_patch.html#normalize_patch_transforms',
                                                                                               'fastMONAI/vision_patch.py'),
                                        'fastMONAI.vision_patch.patch_inference': ( 'vision_patch.html#patch_inference',
                                                                                    'fastMONAI/vision_patch.py')},
            'fastMONAI.vision_plot': { 'fastMONAI.vision_plot._get_slice': ('vision_plot.html#_get_slice', 'fastMONAI/vision_plot.py'),
                                       'fastMONAI.vision_plot.find_max_slice': ( 'vision_plot.html#find_max_slice',
                                                                                 'fastMONAI/vision_plot.py'),
                                       'fastMONAI.vision_plot.show_med_img': ('vision_plot.html#show_med_img', 'fastMONAI/vision_plot.py'),
                                       'fastMONAI.vision_plot.show_segmentation_comparison': ( 'vision_plot.html#show_segmentation_comparison',
                                                                                               'fastMONAI/vision_plot.py'),
                                       'fastMONAI.vision_plot.validate_anatomical_plane': ( 'vision_plot.html#validate_anatomical_plane',
                                                                                            'fastMONAI/vision_plot.py')}}}



#imports all of the functions and classes from the fastMONAI library
from .vision_core import *
from .vision_data import *
from .vision_augmentation import *
from .vision_loss import *
from .vision_metrics import *
from .vision_inference import *
from .vision_patch import *
from .utils import *
from .external_data import *
from .dataset_info import *

from fastai.vision.all import *


# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/03_vision_augment.ipynb.

# %% auto #0
__all__ = ['CustomDictTransform', 'do_pad_or_crop', 'PadOrCrop', 'ZNormalization', 'RescaleIntensity', 'NormalizeIntensity',
           'BraTSMaskConverter', 'BinaryConverter', 'RandomGhosting', 'RandomSpike', 'RandomNoise', 'RandomBiasField',
           'RandomBlur', 'RandomGamma', 'RandomIntensityScale', 'RandomMotion', 'RandomAnisotropy', 'RandomCutout',
           'RandomElasticDeformation', 'RandomAffine', 'RandomFlip', 'OneOf', 'GpuPatchAugmentation',
           'gpu_patch_augmentations', 'suggest_patch_augmentations']

# %% ../nbs/03_vision_augment.ipynb #2d6694aa
from fastai.data.all import *
from .vision_core import *
import torch.nn.functional as F
import math
import torchio as tio
from monai.transforms import NormalizeIntensity as MonaiNormalizeIntensity

# %% ../nbs/03_vision_augment.ipynb #7c2bacd9
class CustomDictTransform(ItemTransform):
    """A class that serves as a wrapper to perform an identical transformation on both
    the image and the target (if it's a mask).
    """

    split_idx = 0  # Only perform transformations on training data. Use TTA() for transformations on validation data.

    def __init__(self, aug):
        """Constructs CustomDictTransform object.

        Args:
            aug (Callable): Function to apply augmentation on the image.
        """
        self.aug = aug

    @property
    def tio_transform(self):
        """Return the underlying TorchIO transform.

        This property enables using fastMONAI wrappers in patch-based workflows
        where raw TorchIO transforms are needed for tio.Compose().
        """
        return self.aug

    def encodes(self, x):
        """
        Applies the stored transformation to an image, and the same random transformation
        to the target if it is a mask. If the target is not a mask, it returns the target as is.

        Args:
            x (Tuple[MedImage, Union[MedMask, TensorCategory]]): A tuple containing the
            image and the target.

        Returns:
            Tuple[MedImage, Union[MedMask, TensorCategory]]: The transformed image and target.
            If the target is a mask, it's transformed identically to the image. If the target
            is not a mask, the original target is returned.
        """
        img, y_true = x
        
        # Use identity affine if MedImage.affine_matrix is not set
        affine = MedImage.affine_matrix if MedImage.affine_matrix is not None else np.eye(4)

        if isinstance(y_true, (MedMask)):
            aug = self.aug(tio.Subject(img=tio.ScalarImage(tensor=img, affine=affine),
                                        mask=tio.LabelMap(tensor=y_true, affine=affine)))
            return MedImage.create(aug['img'].data), MedMask.create(aug['mask'].data)

        aug = self.aug(tio.Subject(img=tio.ScalarImage(tensor=img)))
        return MedImage.create(aug['img'].data), y_true

# %% ../nbs/03_vision_augment.ipynb #d7e1fdf2
def do_pad_or_crop(o, target_shape, padding_mode, mask_name, dtype=torch.Tensor):
    #TODO:refactorize
    pad_or_crop = tio.CropOrPad(target_shape=target_shape, padding_mode=padding_mode, mask_name=mask_name)
    return dtype(pad_or_crop(o))

# %% ../nbs/03_vision_augment.ipynb #9c75bb84
class PadOrCrop(DisplayedTransform):
    """Resize image using TorchIO `CropOrPad`."""
    
    order = 0

    def __init__(self, size, padding_mode=0, mask_name=None):
        if not is_listy(size): 
            size = [size, size, size]
        self.pad_or_crop = tio.CropOrPad(target_shape=size,
                                    padding_mode=padding_mode, 
                                    mask_name=mask_name)

    @property
    def tio_transform(self):
        """Return the underlying TorchIO transform."""
        return self.pad_or_crop

    def encodes(self, o: (MedImage, MedMask)):
        return type(o)(self.pad_or_crop(o))

# %% ../nbs/03_vision_augment.ipynb #534509q2nn
def _foreground_masking(tensor):
    """Mask for non-zero voxels (nnU-Net-style foreground normalization)."""
    return tensor > 0

# %% ../nbs/03_vision_augment.ipynb #ca95a690
class ZNormalization(DisplayedTransform):
    """Apply TorchIO `ZNormalization`."""

    order = 0

    def __init__(self, masking_method=None, channel_wise=True):
        if masking_method == 'foreground':
            masking_method = _foreground_masking
        self.z_normalization = tio.ZNormalization(masking_method=masking_method)
        self.channel_wise = channel_wise

    @property
    def tio_transform(self):
        """Return the underlying TorchIO transform."""
        return self.z_normalization

    def encodes(self, o: MedImage):
        try:
            if self.channel_wise:
                o = torch.stack([self.z_normalization(c[None])[0] for c in o])
            else: 
                o = self.z_normalization(o)
        except RuntimeError as e:
            if "Standard deviation is 0" in str(e):
                # Calculate mean for debugging information
                mean = float(o.mean())
                
                error_msg = (
                    f"Standard deviation is 0 for image (mean={mean:.3f}).\n"
                    f"This indicates uniform pixel values.\n\n"
                    f"Possible causes:\n"
                    f"• Corrupted or blank image\n"
                    f"• Oversaturated regions\n"  
                    f"• Background-only regions\n"
                    f"• All-zero mask being processed as image\n\n"
                    f"Suggested solutions:\n"
                    f"• Check image quality and acquisition\n"
                    f"• Verify image vs mask data loading"
                )
                raise RuntimeError(error_msg) from e

        return MedImage.create(o)

    def encodes(self, o: MedMask):
        return o

# %% ../nbs/03_vision_augment.ipynb #ee9cf2a3
class RescaleIntensity(DisplayedTransform):
    """Apply TorchIO RescaleIntensity for robust intensity scaling.
    
    Args:
        out_min_max (tuple[float, float]): Output intensity range (min, max)
        in_min_max (tuple[float, float]): Input intensity range (min, max) 
    
    Example for CT images:
        # Normalize CT from air (-1000 HU) to bone (1000 HU) into range (-1, 1)
        transform = RescaleIntensity(out_min_max=(-1, 1), in_min_max=(-1000, 1000))
    """
    
    order = 0
    
    def __init__(self, out_min_max: tuple[float, float], in_min_max: tuple[float, float]):
        self.rescale = tio.RescaleIntensity(out_min_max=out_min_max, in_min_max=in_min_max)

    @property
    def tio_transform(self):
        """Return the underlying TorchIO transform."""
        return self.rescale

    def encodes(self, o: MedImage):
        return MedImage.create(self.rescale(o))
        
    def encodes(self, o: MedMask):
        return o

# %% ../nbs/03_vision_augment.ipynb #341dc45f
class _TioNormalizeIntensity(tio.IntensityTransform):
    """TorchIO-compatible wrapper for MONAI NormalizeIntensity.

    Enables NormalizeIntensity to work in patch-based workflows
    where raw TorchIO transforms are needed for tio.Compose().
    """

    def __init__(self, nonzero=True, channel_wise=True,
                 subtrahend=None, divisor=None, **kwargs):
        super().__init__(p=1, **kwargs)
        self.channel_wise = channel_wise
        self.transform = MonaiNormalizeIntensity(
            nonzero=nonzero,
            channel_wise=False,
            subtrahend=subtrahend,
            divisor=divisor
        )

    def apply_transform(self, subject):
        for image in self.get_images(subject):
            data = image.data
            if self.channel_wise:
                result = torch.stack([self.transform(c[None])[0] for c in data])
            else:
                result = torch.Tensor(self.transform(data))
            image.set_data(result)
        return subject


class NormalizeIntensity(DisplayedTransform):
    """Apply MONAI NormalizeIntensity.
    
    Args:
        nonzero (bool): Only normalize non-zero values (default: True)
        channel_wise (bool): Apply normalization per channel (default: True)
        subtrahend (float, optional): Value to subtract  
        divisor (float, optional): Value to divide by
    """
    
    order = 0
    
    def __init__(self, nonzero: bool = True, channel_wise: bool = True, 
                 subtrahend: float = None, divisor: float = None):
        self.nonzero = nonzero
        self.channel_wise = channel_wise
        self.subtrahend = subtrahend
        self.divisor = divisor
        
        self.transform = MonaiNormalizeIntensity(
            nonzero=nonzero,
            channel_wise=False,  # Always 'False', we handle channel-wise manually
            subtrahend=subtrahend,
            divisor=divisor
        )
        self._tio_normalize = _TioNormalizeIntensity(
            nonzero=nonzero,
            channel_wise=channel_wise,
            subtrahend=subtrahend,
            divisor=divisor
        )

    @property
    def tio_transform(self):
        """Return TorchIO-compatible transform for patch-based workflows."""
        return self._tio_normalize
    
    def encodes(self, o: MedImage):
        if self.channel_wise:
            result = torch.stack([self.transform(c[None])[0] for c in o])
        else:
            result = torch.Tensor(self.transform(o))
        
        return MedImage.create(result)
    
    def encodes(self, o: MedMask):
        return o

# %% ../nbs/03_vision_augment.ipynb #be431cc2
class BraTSMaskConverter(DisplayedTransform):
    '''Convert BraTS masks.'''

    order=1

    def encodes(self, o:(MedImage)): return o

    def encodes(self, o:(MedMask)):
        o = torch.where(o==4, 3., o)
        return MedMask.create(o)

# %% ../nbs/03_vision_augment.ipynb #f63701cb
class BinaryConverter(DisplayedTransform):
    '''Convert to binary mask.'''

    order=1

    def encodes(self, o: MedImage): 
        return o

    def encodes(self, o: MedMask):
        o = torch.where(o>0, 1., 0)
        return MedMask.create(o)

# %% ../nbs/03_vision_augment.ipynb #6b5a795f
class RandomGhosting(DisplayedTransform):
    """Apply TorchIO `RandomGhosting`."""
    
    split_idx, order = 0, 1

    def __init__(self, intensity=(0.5, 1), p=0.5):
        self.add_ghosts = tio.RandomGhosting(intensity=intensity, p=p)

    @property
    def tio_transform(self):
        """Return the underlying TorchIO transform."""
        return self.add_ghosts

    def encodes(self, o: MedImage):
        return MedImage.create(self.add_ghosts(o))

    def encodes(self, o: MedMask):
        return o

# %% ../nbs/03_vision_augment.ipynb #bf9ce6fc
class RandomSpike(DisplayedTransform):
    '''Apply TorchIO `RandomSpike`.'''
    
    split_idx, order = 0, 1

    def __init__(self, num_spikes=1, intensity=(1, 3), p=0.5):
        self.add_spikes = tio.RandomSpike(num_spikes=num_spikes, intensity=intensity, p=p)

    @property
    def tio_transform(self):
        """Return the underlying TorchIO transform."""
        return self.add_spikes

    def encodes(self, o: MedImage): 
        return MedImage.create(self.add_spikes(o))
        
    def encodes(self, o: MedMask):
        return o

# %% ../nbs/03_vision_augment.ipynb #a48e315a
class RandomNoise(DisplayedTransform):
    '''Apply TorchIO `RandomNoise`.'''

    split_idx, order = 0, 1

    def __init__(self, mean=0, std=(0, 0.25), p=0.5):
        self.add_noise = tio.RandomNoise(mean=mean, std=std, p=p)

    @property
    def tio_transform(self):
        """Return the underlying TorchIO transform."""
        return self.add_noise

    def encodes(self, o: MedImage): 
        return MedImage.create(self.add_noise(o))
    
    def encodes(self, o: MedMask):
        return o

# %% ../nbs/03_vision_augment.ipynb #6623a796
class RandomBiasField(DisplayedTransform):
    '''Apply TorchIO `RandomBiasField`.'''

    split_idx, order = 0, 1

    def __init__(self, coefficients=0.5, order=3, p=0.5):
        self.add_biasfield = tio.RandomBiasField(coefficients=coefficients, order=order, p=p)

    @property
    def tio_transform(self):
        """Return the underlying TorchIO transform."""
        return self.add_biasfield

    def encodes(self, o: MedImage): 
        return MedImage.create(self.add_biasfield(o))
        
    def encodes(self, o: MedMask):
        return o

# %% ../nbs/03_vision_augment.ipynb #7e7d29c1
class RandomBlur(DisplayedTransform):
    '''Apply TorchIO `RandomBlur`.'''

    split_idx, order = 0, 1

    def __init__(self, std=(0, 2), p=0.5):
        self.add_blur = tio.RandomBlur(std=std, p=p)

    @property
    def tio_transform(self):
        """Return the underlying TorchIO transform."""
        return self.add_blur

    def encodes(self, o: MedImage): 
        return MedImage.create(self.add_blur(o))
    
    def encodes(self, o: MedMask):
        return o

# %% ../nbs/03_vision_augment.ipynb #cd45c2c9
class RandomGamma(DisplayedTransform):
    '''Apply TorchIO `RandomGamma`.'''

    split_idx, order = 0, 1

    def __init__(self, log_gamma=(-0.3, 0.3), p=0.5):
        self.add_gamma = tio.RandomGamma(log_gamma=log_gamma, p=p)

    @property
    def tio_transform(self):
        """Return the underlying TorchIO transform."""
        return self.add_gamma

    def encodes(self, o: MedImage): 
        return MedImage.create(self.add_gamma(o))
    
    def encodes(self, o: MedMask):
        return o

# %% ../nbs/03_vision_augment.ipynb #4116aa33
class _TioRandomIntensityScale(tio.IntensityTransform):
    """TorchIO-compatible RandomIntensityScale for patch-based workflows.

    Randomly scales image intensities by a multiplicative factor.
    Only applies to ScalarImage keys (not LabelMap), which is the standard
    TorchIO IntensityTransform behavior.
    """

    def __init__(self, scale_range=(0.5, 2.0), p=0.5, **kwargs):
        super().__init__(p=p, **kwargs)
        self.scale_range = scale_range

    def apply_transform(self, subject):
        for image in self.get_images(subject):
            scale = torch.empty(1).uniform_(self.scale_range[0], self.scale_range[1]).item()
            image.set_data(image.data * scale)
        return subject


class RandomIntensityScale(DisplayedTransform):
    """Randomly scale image intensities by a multiplicative factor.

    Useful for domain generalization across different acquisition protocols
    with varying intensity ranges.

    Args:
        scale_range (tuple[float, float]): Range of scale factors (min, max).
            Values > 1 increase intensity, < 1 decrease intensity.
        p (float): Probability of applying the transform (default: 0.5)

    Example:
        # Scale intensities randomly between 0.5x and 2.0x
        transform = RandomIntensityScale(scale_range=(0.5, 2.0), p=0.3)
    """

    split_idx, order = 0, 1

    def __init__(self, scale_range: tuple[float, float] = (0.5, 2.0), p: float = 0.5):
        self.scale_range = scale_range
        self.p = p
        self._tio_intensity_scale = _TioRandomIntensityScale(
            scale_range=scale_range, p=p
        )

    @property
    def tio_transform(self):
        """Return TorchIO-compatible transform for patch-based workflows."""
        return self._tio_intensity_scale

    def encodes(self, o: MedImage):
        if torch.rand(1).item() > self.p:
            return o
        scale = torch.empty(1).uniform_(self.scale_range[0], self.scale_range[1]).item()
        return MedImage.create(o * scale)

    def encodes(self, o: MedMask):
        return o

# %% ../nbs/03_vision_augment.ipynb #bf89ffbe
class RandomMotion(DisplayedTransform):
    """Apply TorchIO `RandomMotion`."""

    split_idx, order = 0, 1

    def __init__(
            self, 
            degrees=10, 
            translation=10, 
            num_transforms=2, 
            image_interpolation='linear', 
            p=0.5
        ):
        self.add_motion = tio.RandomMotion(
            degrees=degrees, 
            translation=translation, 
            num_transforms=num_transforms, 
            image_interpolation=image_interpolation, 
            p=p
        )

    @property
    def tio_transform(self):
        """Return the underlying TorchIO transform."""
        return self.add_motion

    def encodes(self, o: MedImage):
        return MedImage.create(self.add_motion(o))

    def encodes(self, o: MedMask):
        return o

# %% ../nbs/03_vision_augment.ipynb #cl3ei8hm3z9
class RandomAnisotropy(DisplayedTransform):
    '''Apply TorchIO `RandomAnisotropy`.'''

    split_idx, order = 0, 1

    def __init__(self, axes=(0, 1, 2), downsampling=(1.5, 5),
                 image_interpolation='linear', scalars_only=True, p=0.5):
        self.add_anisotropy = tio.RandomAnisotropy(
            axes=axes, downsampling=downsampling,
            image_interpolation=image_interpolation,
            scalars_only=scalars_only, p=p)

    @property
    def tio_transform(self):
        """Return the underlying TorchIO transform."""
        return self.add_anisotropy

    def encodes(self, o: MedImage):
        return MedImage.create(self.add_anisotropy(o))

    def encodes(self, o: MedMask):
        return o

# %% ../nbs/03_vision_augment.ipynb #e7ea6486
def _create_ellipsoid_mask(shape, center, radii):
    """Create a 3D ellipsoid mask.

    Args:
        shape: (D, H, W) shape of the volume
        center: (z, y, x) center of ellipsoid
        radii: (rz, ry, rx) radii along each axis

    Returns:
        Boolean mask where True = inside ellipsoid
    """
    z, y, x = torch.meshgrid(
        torch.arange(shape[0]),
        torch.arange(shape[1]),
        torch.arange(shape[2]),
        indexing='ij'
    )
    dist = ((z - center[0]) / radii[0]) ** 2 + \
           ((y - center[1]) / radii[1]) ** 2 + \
           ((x - center[2]) / radii[2]) ** 2
    return dist <= 1.0

# %% ../nbs/03_vision_augment.ipynb #935c13e7
class _TioRandomCutout(tio.IntensityTransform):
    """TorchIO-compatible RandomCutout for patch-based workflows.

    When mask_only=True, cutouts only affect voxels where the mask is positive.
    The mask should be available in the Subject as 'mask' key.
    """

    def __init__(self, holes=1, spatial_size=8, fill_value=None,
                 max_holes=None, max_spatial_size=None, mask_only=True, p=0.2, **kwargs):
        super().__init__(p=p, **kwargs)
        self.holes = holes
        self.spatial_size = spatial_size
        self.fill_value = fill_value
        self.max_holes = max_holes
        self.max_spatial_size = max_spatial_size
        self.mask_only = mask_only

    def _apply_cutout(self, data, fill_val, mask_tensor=None):
        """Apply spherical cutout(s) to a tensor.

        Args:
            data: Input tensor of shape (C, D, H, W)
            fill_val: Value to fill cutout regions
            mask_tensor: Optional mask tensor for mask-only cutouts

        Returns:
            Tensor with cutout applied
        """
        result = data.clone()
        n_holes = torch.randint(self.holes, (self.max_holes or self.holes) + 1, (1,)).item()

        spatial_shape = data.shape[1:]  # (D, H, W)
        min_size = self.spatial_size if isinstance(self.spatial_size, int) else self.spatial_size[0]
        max_size = self.max_spatial_size or self.spatial_size
        max_size = max_size if isinstance(max_size, int) else max_size[0]

        for _ in range(n_holes):
            # Random size for this hole
            size = torch.randint(min_size, max_size + 1, (3,))
            radii = size.float() / 2

            # Random center (ensure hole fits in volume)
            center = [
                torch.randint(int(radii[i].item()),
                              max(spatial_shape[i] - int(radii[i].item()), int(radii[i].item()) + 1),
                              (1,)).item()
                for i in range(3)
            ]

            ellipsoid = _create_ellipsoid_mask(spatial_shape, center, radii)

            if self.mask_only and mask_tensor is not None:
                # INTERSECT with tumor mask - only affect tumor voxels
                tumor_mask = mask_tensor[0] > 0
                cutout_region = ellipsoid & tumor_mask
            else:
                cutout_region = ellipsoid

            result[:, cutout_region] = fill_val

        return result

    def apply_transform(self, subject):
        # Get mask if available for mask-only cutouts
        mask_tensor = None
        if self.mask_only and 'mask' in subject:
            mask_tensor = subject['mask'].data

            # Skip if mask is empty
            if mask_tensor is not None and not (mask_tensor > 0).any():
                return subject

        for image in self.get_images(subject):
            data = image.data
            fill_val = self.fill_value if self.fill_value is not None else float(data.min())
            result = self._apply_cutout(data, fill_val, mask_tensor)
            image.set_data(result)
        return subject

# %% ../nbs/03_vision_augment.ipynb #bc99f04d
class RandomCutout(ItemTransform):
    """Randomly erase spherical regions in 3D medical images with mask-aware placement.

    Simulates post-operative surgical cavities by filling random ellipsoid
    volumes with specified values. When mask_only=True (default), cutouts only
    affect voxels inside the segmentation mask, ensuring no healthy tissue is modified.

    Args:
        holes: Minimum number of cutout regions. Default: 1.
        max_holes: Maximum number of regions. Default: 3.
        spatial_size: Minimum cutout diameter in voxels. Default: 8.
        max_spatial_size: Maximum cutout diameter. Default: 16.
        fill: Fill value - 'min', 'mean', 'random', or float. Default: 'min'.
        mask_only: If True, cutouts only affect mask-positive voxels (tumor tissue).
            If False, cutouts can affect any voxel (original behavior). Default: True.
        p: Probability of applying transform. Default: 0.2.

    Example:
        >>> # Simulate post-op cavities only within tumor regions
        >>> tfm = RandomCutout(holes=1, max_holes=2, spatial_size=10,
        ...                    max_spatial_size=25, fill='min', mask_only=True, p=0.2)

        >>> # Original behavior - cutouts anywhere in the volume
        >>> tfm = RandomCutout(mask_only=False, p=0.2)
    """

    split_idx, order = 0, 1

    def __init__(self, holes=1, max_holes=3, spatial_size=8,
                 max_spatial_size=16, fill='min', mask_only=True, p=0.2):
        self.holes = holes
        self.max_holes = max_holes
        self.spatial_size = spatial_size
        self.max_spatial_size = max_spatial_size
        self.fill = fill
        self.mask_only = mask_only
        self.p = p

        self._tio_cutout = _TioRandomCutout(
            holes=holes, spatial_size=spatial_size,
            fill_value=None if isinstance(fill, str) else fill,
            max_holes=max_holes, max_spatial_size=max_spatial_size,
            mask_only=mask_only, p=p
        )

    @property
    def tio_transform(self):
        """Return TorchIO-compatible transform for patch-based workflows.

        Note: For mask-aware cutouts in patch workflows, the mask must be
        available in the TorchIO Subject as 'mask' key.
        """
        return self._tio_cutout

    def _get_fill_value(self, tensor):
        if self.fill == 'min': return float(tensor.min())
        elif self.fill == 'mean': return float(tensor.mean())
        elif self.fill == 'random':
            return torch.empty(1).uniform_(float(tensor.min()), float(tensor.max())).item()
        else: return self.fill

    def encodes(self, x):
        """Apply mask-aware cutout to image.

        Args:
            x: Tuple of (MedImage, target) where target is MedMask or TensorCategory

        Returns:
            Tuple of (transformed MedImage, unchanged target)
        """
        img, y_true = x

        # Probability check
        if torch.rand(1).item() > self.p:
            return img, y_true

        # Get mask data if available (as numpy for safe boolean operations)
        mask_np = None
        tumor_coords = None
        if isinstance(y_true, MedMask):
            mask_np = y_true.numpy()
            # Get coordinates of tumor voxels for mask-aware center placement
            if self.mask_only and (mask_np > 0).any():
                tumor_coords = np.argwhere(mask_np[0] > 0)  # Shape: (N, 3) for z, y, x

        # Skip cutout if mask_only=True but no mask or empty mask
        if self.mask_only:
            if mask_np is None or tumor_coords is None or len(tumor_coords) == 0:
                return img, y_true

        # Work with numpy array to avoid tensor subclass issues
        result_np = img.numpy().copy()
        spatial_shape = img.shape[1:]  # (D, H, W)
        fill_val = self._get_fill_value(img)

        n_holes = torch.randint(self.holes, self.max_holes + 1, (1,)).item()

        min_size = self.spatial_size if isinstance(self.spatial_size, int) else self.spatial_size[0]
        max_size = self.max_spatial_size or self.spatial_size
        max_size = max_size if isinstance(max_size, int) else max_size[0]

        for _ in range(n_holes):
            # Random size for this hole
            size = torch.randint(min_size, max_size + 1, (3,))
            radii = size.float() / 2

            if self.mask_only and tumor_coords is not None:
                # Pick center from within tumor region to ensure intersection
                idx = torch.randint(0, len(tumor_coords), (1,)).item()
                center = tumor_coords[idx].tolist()  # [z, y, x]
            else:
                # Random center anywhere (ensure hole fits in volume)
                center = [
                    torch.randint(int(radii[i].item()),
                                  max(spatial_shape[i] - int(radii[i].item()), int(radii[i].item()) + 1),
                                  (1,)).item()
                    for i in range(3)
                ]

            # Create ellipsoid mask (numpy)
            ellipsoid = _create_ellipsoid_mask(spatial_shape, center, radii).numpy()

            if self.mask_only and mask_np is not None:
                # INTERSECT with tumor mask - only affect tumor voxels
                tumor_mask = mask_np[0] > 0
                cutout_region = ellipsoid & tumor_mask
            else:
                cutout_region = ellipsoid

            result_np[:, cutout_region] = fill_val

        return MedImage.create(torch.from_numpy(result_np)), y_true

# %% ../nbs/03_vision_augment.ipynb #3c4f113d
class RandomElasticDeformation(CustomDictTransform):
    """Apply TorchIO `RandomElasticDeformation`."""

    def __init__(self, num_control_points=7, max_displacement=7.5,
                 image_interpolation='linear', p=0.5):
                     
        super().__init__(tio.RandomElasticDeformation(
            num_control_points=num_control_points,
            max_displacement=max_displacement,
            image_interpolation=image_interpolation,
            p=p))

# %% ../nbs/03_vision_augment.ipynb #feaecd21
class RandomAffine(CustomDictTransform):
    """Apply TorchIO `RandomAffine`."""

    def __init__(self, scales=0, degrees=10, translation=0, isotropic=False,
                 image_interpolation='linear', default_pad_value=0., p=0.5):
                     
        super().__init__(tio.RandomAffine(
            scales=scales,
            degrees=degrees,
            translation=translation,
            isotropic=isotropic,
            image_interpolation=image_interpolation,
            default_pad_value=default_pad_value,
            p=p))

# %% ../nbs/03_vision_augment.ipynb #022c90cf
class RandomFlip(CustomDictTransform):
    """Apply TorchIO `RandomFlip`."""

    def __init__(self, axes='LRAPIS', p=0.5):
        super().__init__(tio.RandomFlip(axes=axes, flip_probability=p))

# %% ../nbs/03_vision_augment.ipynb #ddd7b99b
class OneOf(CustomDictTransform):
    """Apply only one of the given transforms using TorchIO `OneOf`."""

    def __init__(self, transform_dict, p=1):
        super().__init__(tio.OneOf(transform_dict, p=p))

# %% ../nbs/03_vision_augment.ipynb #lqet5pabzy
def _compute_patch_aug_params(patch_size, target_spacing,
                              anisotropy_threshold=3.0,
                              translation_fraction=0.15):
    """Compute geometry-aware augmentation parameters from patch/spacing metadata.

    Shared logic used by both suggest_patch_augmentations (CPU/TorchIO) and
    gpu_patch_augmentations (GPU-batched). Extracts rotation degrees, translation
    offsets, and RandomAnisotropy axes from the spatial configuration.

    Anisotropy detection: if max(spacing)/min(spacing) >= threshold, rotation
    is restricted to 5 deg out-of-plane and 30 deg in-plane. Otherwise 30 deg
    symmetric.

    Args:
        patch_size: List/tuple of 3 ints -- patch dimensions.
        target_spacing: List/tuple of 3 floats -- voxel spacing.
        anisotropy_threshold: Ratio threshold for anisotropy detection (default 3.0).
        translation_fraction: Fraction of patch_size for translation (default 0.15).

    Returns:
        dict with keys:
            'degrees': tuple of 3 ints (per-axis max rotation in degrees)
            'translation': tuple of 3 ints (per-axis translation in voxels)
            'aniso_axes': tuple of ints (axes where patch_size > 1)
            'is_aniso': bool (whether spacing is anisotropic)
    """
    if len(patch_size) != 3:
        raise ValueError(f"patch_size must have 3 elements, got {len(patch_size)}")
    if len(target_spacing) != 3:
        raise ValueError(f"target_spacing must have 3 elements, got {len(target_spacing)}")

    spacing = list(target_spacing)
    ratio = max(spacing) / min(spacing)
    is_aniso = ratio >= anisotropy_threshold
    aniso_axis = spacing.index(max(spacing)) if is_aniso else None

    if is_aniso:
        degrees = [5, 5, 5]
        degrees[aniso_axis] = 30
        degrees = tuple(degrees)
    else:
        degrees = (30, 30, 30)

    translation = tuple(round(p * translation_fraction) for p in patch_size)
    aniso_axes = tuple(i for i in range(3) if patch_size[i] > 1)

    return {
        'degrees': degrees,
        'translation': translation,
        'aniso_axes': aniso_axes,
        'is_aniso': is_aniso,
    }

# %% ../nbs/03_vision_augment.ipynb #oef9rtzvbw
def _build_rotation_matrix_3d(angles_rad):
    """Build [N, 3, 3] rotation matrices from [N, 3] Euler angles (XYZ extrinsic).

    Computes R = Rz @ Ry @ Rx for each sample in the batch.

    Args:
        angles_rad: Tensor of shape [N, 3] with rotation angles in radians
                    for each axis (x, y, z).

    Returns:
        Tensor of shape [N, 3, 3] -- rotation matrices.
    """
    cos = torch.cos(angles_rad)
    sin = torch.sin(angles_rad)
    cx, cy, cz = cos[:, 0], cos[:, 1], cos[:, 2]
    sx, sy, sz = sin[:, 0], sin[:, 1], sin[:, 2]

    # R = Rz @ Ry @ Rx (combined formula)
    r00 = cy * cz;          r01 = sx * sy * cz - cx * sz;  r02 = cx * sy * cz + sx * sz
    r10 = cy * sz;          r11 = sx * sy * sz + cx * cz;  r12 = cx * sy * sz - sx * cz
    r20 = -sy;              r21 = sx * cy;                  r22 = cx * cy

    R = torch.stack([
        torch.stack([r00, r01, r02], dim=-1),
        torch.stack([r10, r11, r12], dim=-1),
        torch.stack([r20, r21, r22], dim=-1),
    ], dim=-2)
    return R

# %% ../nbs/03_vision_augment.ipynb #ts2kolv83z
class GpuPatchAugmentation:
    """GPU-batched augmentation for patch-based training.

    Operates on [B, C, D, H, W] tensors already on GPU. All operations run
    under torch.no_grad() since augmentation does not need gradient tracking.

    Transform order: spatial (affine, anisotropy, flip) then intensity
    (gamma, intensity_scale, noise, blur). Spatial transforms apply the
    same parameters to both image and mask. Intensity transforms skip the mask.

    Each transform is controlled by a parameter dict with at minimum a 'p' key
    for per-sample probability. Pass None to disable a transform.

    Args:
        affine: dict with keys 'scales', 'degrees', 'translation',
                'default_pad_value', 'p'. None to disable.
        anisotropy: dict with keys 'axes', 'downsampling', 'p'. None to disable.
        flip: dict with keys 'axes', 'p'. None to disable.
        gamma: dict with keys 'log_gamma', 'p'. None to disable.
        intensity_scale: dict with keys 'scale_range', 'p'. None to disable.
        noise: dict with keys 'std', 'p'. None to disable.
        blur: dict with keys 'std', 'p'. None to disable.

    Example::

        >>> gpu_aug = GpuPatchAugmentation(
        ...     affine={'scales': (0.7, 1.4), 'degrees': (30, 30, 30),
        ...             'translation': (25, 25, 10), 'default_pad_value': 0., 'p': 0.2},
        ...     gamma={'log_gamma': (-0.3, 0.3), 'p': 0.3},
        ...     flip={'axes': (0, 1, 2), 'p': 0.5},
        ... )
        >>> img_aug, mask_aug = gpu_aug(img_gpu, mask_gpu)
    """

    def __init__(self, affine=None, anisotropy=None, flip=None,
                 gamma=None, intensity_scale=None, noise=None, blur=None):
        self.affine = affine
        self.anisotropy = anisotropy
        self.flip = flip
        self.gamma = gamma
        self.intensity_scale = intensity_scale
        self.noise = noise
        self.blur = blur

    def __call__(self, img, mask=None):
        """Apply GPU augmentations to a batch.

        Args:
            img: Tensor [B, C, D, H, W] (float).
            mask: Tensor [B, C, D, H, W] (float), or None.

        Returns:
            Tuple (img, mask). mask is None if input was None.
        """
        with torch.no_grad():
            # Spatial transforms (same params for img and mask)
            if self.affine is not None:
                img, mask = self._apply_affine(img, mask)
            if self.anisotropy is not None:
                img, mask = self._apply_anisotropy(img, mask)
            if self.flip is not None:
                img, mask = self._apply_flip(img, mask)
            # Intensity transforms (img only)
            if self.gamma is not None:
                img = self._apply_gamma(img)
            if self.intensity_scale is not None:
                img = self._apply_intensity_scale(img)
            if self.noise is not None:
                img = self._apply_noise(img)
            if self.blur is not None:
                img = self._apply_blur(img)
        return img, mask

    def _apply_affine(self, img, mask):
        """Batched random affine via F.affine_grid + F.grid_sample."""
        cfg = self.affine
        B = img.shape[0]
        device = img.device
        dtype = img.dtype

        # Per-sample probability
        do_tfm = torch.rand(B, device=device) < cfg['p']
        if not do_tfm.any():
            return img, mask

        # Start with identity theta [B, 3, 4]
        theta = torch.zeros(B, 3, 4, device=device, dtype=dtype)
        theta[:, 0, 0] = 1.0
        theta[:, 1, 1] = 1.0
        theta[:, 2, 2] = 1.0

        idx = do_tfm.nonzero(as_tuple=True)[0]
        n = idx.shape[0]

        # Random scales per axis
        s_lo, s_hi = cfg['scales']
        scales = torch.empty(n, 3, device=device, dtype=dtype).uniform_(s_lo, s_hi)

        # Random rotation angles (degrees -> radians)
        degrees = cfg['degrees']
        if not isinstance(degrees, (list, tuple)):
            degrees = (degrees, degrees, degrees)
        angles_deg = torch.stack([
            torch.empty(n, device=device, dtype=dtype).uniform_(-degrees[0], degrees[0]),
            torch.empty(n, device=device, dtype=dtype).uniform_(-degrees[1], degrees[1]),
            torch.empty(n, device=device, dtype=dtype).uniform_(-degrees[2], degrees[2]),
        ], dim=1)  # [n, 3]
        angles_rad = angles_deg * (math.pi / 180.0)

        # Build rotation matrices [n, 3, 3]
        R = _build_rotation_matrix_3d(angles_rad)

        # Scale matrix: S @ R -> [n, 3, 3]
        S = torch.diag_embed(scales)  # [n, 3, 3]
        SR = S @ R  # [n, 3, 3]

        # Random translation (voxels -> normalized [-1, 1] coords)
        translation = cfg['translation']
        if not isinstance(translation, (list, tuple)):
            translation = (translation, translation, translation)
        spatial_size = img.shape[2:]  # (D, H, W)
        t_norm = torch.stack([
            torch.empty(n, device=device, dtype=dtype).uniform_(
                -translation[i], translation[i]
            ) * 2.0 / spatial_size[i]
            for i in range(3)
        ], dim=1)  # [n, 3]

        # Assemble theta for active samples
        theta[idx, :3, :3] = SR
        theta[idx, :3, 3] = t_norm

        # Apply grid_sample
        grid = F.affine_grid(theta, img.shape, align_corners=False)
        pad_val = cfg.get('default_pad_value', 0.)
        # For non-zero padding, shift values, sample, shift back
        if pad_val != 0.:
            img = img - pad_val
        img = F.grid_sample(img, grid, mode='bilinear',
                            padding_mode='zeros', align_corners=False)
        if pad_val != 0.:
            img = img + pad_val

        if mask is not None:
            mask = F.grid_sample(mask.float(), grid, mode='nearest',
                                 padding_mode='zeros', align_corners=False)
        return img, mask

    def _apply_anisotropy(self, img, mask):
        """Per-sample anisotropy simulation via F.interpolate.

        Downsample along a random axis with nearest interpolation,
        then upsample back with trilinear (matches TorchIO behavior).
        Only affects img, not mask (anisotropy is intensity degradation).
        """
        cfg = self.anisotropy
        B = img.shape[0]
        device = img.device
        ds_lo, ds_hi = cfg['downsampling']
        axes = cfg['axes']

        for i in range(B):
            if torch.rand(1, device=device).item() >= cfg['p']:
                continue
            # Pick random axis and downsampling factor
            axis_idx = torch.randint(len(axes), (1,), device=device).item()
            axis = axes[axis_idx]
            factor = torch.empty(1, device=device).uniform_(ds_lo, ds_hi).item()

            sample = img[i:i+1]  # [1, C, D, H, W]
            orig_size = list(sample.shape[2:])
            down_size = list(orig_size)
            down_size[axis] = max(1, round(orig_size[axis] / factor))

            # Downsample with nearest, upsample with trilinear
            down = F.interpolate(sample, size=down_size, mode='nearest')
            up = F.interpolate(down, size=orig_size, mode='trilinear',
                               align_corners=False)
            img[i:i+1] = up

        return img, mask

    def _apply_flip(self, img, mask):
        """Per-sample random flip along configured axes."""
        cfg = self.flip
        B = img.shape[0]
        device = img.device
        axes = cfg['axes']
        p = cfg['p']

        for i in range(B):
            # Each axis is independently flipped with probability p
            flip_dims = []
            for axis in axes:
                if torch.rand(1, device=device).item() < p:
                    # axis 0 -> dim 2 (D), axis 1 -> dim 3 (H), axis 2 -> dim 4 (W)
                    # but img[i] is [C, D, H, W], so axis 0 -> dim 1, etc.
                    flip_dims.append(axis + 2)  # +2 for batch and channel dims
            if flip_dims:
                img[i] = torch.flip(img[i], dims=[d - 1 for d in flip_dims])  # -1 since no batch dim
                if mask is not None:
                    mask[i] = torch.flip(mask[i], dims=[d - 1 for d in flip_dims])

        return img, mask

    def _apply_gamma(self, img):
        """Batched gamma correction with per-sample random gamma."""
        cfg = self.gamma
        B = img.shape[0]
        device = img.device
        dtype = img.dtype
        log_lo, log_hi = cfg['log_gamma']

        active = torch.rand(B, device=device) < cfg['p']
        if not active.any():
            return img

        # Only apply clamp + pow to active samples (clamp destroys negatives)
        active_idx = active.nonzero(as_tuple=True)[0]
        log_gamma = torch.empty(active_idx.shape[0], device=device, dtype=dtype).uniform_(log_lo, log_hi)
        gamma = torch.exp(log_gamma).view(-1, 1, 1, 1, 1)
        img[active_idx] = img[active_idx].clamp(min=0).pow(gamma)
        return img

    def _apply_intensity_scale(self, img):
        """Batched intensity scaling with per-sample random factors."""
        cfg = self.intensity_scale
        B = img.shape[0]
        device = img.device
        dtype = img.dtype
        s_lo, s_hi = cfg['scale_range']

        # Per-sample scale (inactive get scale=1)
        scale = torch.empty(B, device=device, dtype=dtype).uniform_(s_lo, s_hi)
        active = torch.rand(B, device=device) < cfg['p']
        scale = torch.where(active, scale, torch.ones_like(scale))

        img = img * scale.view(B, 1, 1, 1, 1)
        return img

    def _apply_noise(self, img):
        """Batched additive Gaussian noise with per-sample random std."""
        cfg = self.noise
        B = img.shape[0]
        device = img.device
        dtype = img.dtype

        std_val = cfg['std']
        if isinstance(std_val, (list, tuple)):
            std_lo, std_hi = std_val
            per_std = torch.empty(B, device=device, dtype=dtype).uniform_(std_lo, std_hi)
        else:
            per_std = torch.full((B,), std_val, device=device, dtype=dtype)

        # Zero std for inactive samples
        active = torch.rand(B, device=device) < cfg['p']
        per_std = torch.where(active, per_std, torch.zeros_like(per_std))

        noise = torch.randn_like(img) * per_std.view(B, 1, 1, 1, 1)
        img = img + noise
        return img

    def _apply_blur(self, img):
        """Batched separable 3D Gaussian blur via F.conv3d with groups trick."""
        cfg = self.blur
        B, C, D, H, W = img.shape
        device = img.device
        dtype = img.dtype

        std_val = cfg['std']
        if isinstance(std_val, (list, tuple)):
            std_lo, std_hi = std_val
        else:
            std_lo, std_hi = 0.0, std_val

        # Per-sample sigma
        sigma = torch.empty(B, device=device, dtype=dtype).uniform_(std_lo, std_hi)
        active = torch.rand(B, device=device) < cfg['p']
        if not active.any():
            return img

        # Fixed kernel size from max sigma
        max_sigma = max(std_hi, 0.01)
        kernel_radius = int(math.ceil(3 * max_sigma))
        kernel_size = 2 * kernel_radius + 1

        # Build per-sample 1D Gaussian kernels [B, kernel_size]
        x = torch.arange(-kernel_radius, kernel_radius + 1,
                          device=device, dtype=dtype)
        # Avoid division by zero for sigma=0
        safe_sigma = torch.where(active, sigma, torch.ones_like(sigma))
        kernels = torch.exp(-x.unsqueeze(0)**2 / (2 * safe_sigma.unsqueeze(1)**2))
        kernels = kernels / kernels.sum(dim=1, keepdim=True)

        # For inactive samples, use delta kernel
        delta = torch.zeros(B, kernel_size, device=device, dtype=dtype)
        delta[:, kernel_radius] = 1.0
        kernels = torch.where(active.unsqueeze(1), kernels, delta)

        # Expand kernels for all channels: [B*C, kernel_size]
        kernels_bc = kernels.unsqueeze(1).expand(B, C, kernel_size).reshape(B * C, kernel_size)

        # Reshape img for grouped convolution: [1, B*C, D, H, W]
        img_grouped = img.reshape(1, B * C, D, H, W)

        # Separable 3D convolution: D-axis, H-axis, W-axis
        pad = kernel_radius

        # D-axis: kernel shape [B*C, 1, K, 1, 1]
        k_d = kernels_bc.reshape(B * C, 1, kernel_size, 1, 1)
        img_grouped = F.pad(img_grouped, (0, 0, 0, 0, pad, pad), mode='replicate')
        img_grouped = F.conv3d(img_grouped, k_d, groups=B * C)

        # H-axis: kernel shape [B*C, 1, 1, K, 1]
        k_h = kernels_bc.reshape(B * C, 1, 1, kernel_size, 1)
        img_grouped = F.pad(img_grouped, (0, 0, pad, pad, 0, 0), mode='replicate')
        img_grouped = F.conv3d(img_grouped, k_h, groups=B * C)

        # W-axis: kernel shape [B*C, 1, 1, 1, K]
        k_w = kernels_bc.reshape(B * C, 1, 1, 1, kernel_size)
        img_grouped = F.pad(img_grouped, (pad, pad, 0, 0, 0, 0), mode='replicate')
        img_grouped = F.conv3d(img_grouped, k_w, groups=B * C)

        return img_grouped.reshape(B, C, D, H, W)

    def __repr__(self):
        parts = []
        for name in ['affine', 'anisotropy', 'flip', 'gamma',
                      'intensity_scale', 'noise', 'blur']:
            cfg = getattr(self, name)
            if cfg is not None:
                parts.append(f"{name}(p={cfg['p']})")
        return f"GpuPatchAugmentation({', '.join(parts)})"

# %% ../nbs/03_vision_augment.ipynb #pdbh1nqo0j7
def gpu_patch_augmentations(patch_size, target_spacing,
                            anisotropy_threshold=3.0,
                            translation_fraction=0.15,
                            affine_p=0.2, anisotropy_p=0.25,
                            gamma_p=0.3, intensity_scale_p=0.1,
                            noise_p=0.1, blur_p=0.2, flip_p=0.5):
    """Create GpuPatchAugmentation with nnU-Net-inspired defaults.

    Factory function that mirrors suggest_patch_augmentations but returns
    a GpuPatchAugmentation for GPU-batched operation. Uses the same shared
    parameter logic via _compute_patch_aug_params.

    Args:
        patch_size: List/tuple of 3 ints -- patch dimensions.
        target_spacing: List/tuple of 3 floats -- voxel spacing.
        anisotropy_threshold: Ratio threshold for anisotropy detection (default 3.0).
        translation_fraction: Fraction of patch_size for translation (default 0.15).
        affine_p: Probability for RandomAffine (default 0.2).
        anisotropy_p: Probability for RandomAnisotropy (default 0.25).
        gamma_p: Probability for RandomGamma (default 0.3).
        intensity_scale_p: Probability for RandomIntensityScale (default 0.1).
        noise_p: Probability for RandomNoise (default 0.1).
        blur_p: Probability for RandomBlur (default 0.2).
        flip_p: Probability for RandomFlip per axis (default 0.5).

    Returns:
        GpuPatchAugmentation instance.

    Example::

        >>> gpu_aug = gpu_patch_augmentations([128, 128, 32], [0.5, 0.5, 1.5])
        >>> dls = MedPatchDataLoaders.from_df(..., gpu_augmentation=gpu_aug)
    """
    params = _compute_patch_aug_params(
        patch_size, target_spacing, anisotropy_threshold, translation_fraction
    )

    affine_cfg = {
        'scales': (0.7, 1.4),
        'degrees': params['degrees'],
        'translation': params['translation'],
        'default_pad_value': 0.,
        'p': affine_p,
    }

    aniso_cfg = None
    if len(params['aniso_axes']) > 0:
        aniso_cfg = {
            'axes': params['aniso_axes'],
            'downsampling': (1.5, 4),
            'p': anisotropy_p,
        }

    return GpuPatchAugmentation(
        affine=affine_cfg,
        anisotropy=aniso_cfg,
        flip={'axes': (0, 1, 2), 'p': flip_p},
        gamma={'log_gamma': (-0.3, 0.3), 'p': gamma_p},
        intensity_scale={'scale_range': (0.75, 1.25), 'p': intensity_scale_p},
        noise={'std': 0.1, 'p': noise_p},
        blur={'std': (0.5, 1.0), 'p': blur_p},
    )

# %% ../nbs/03_vision_augment.ipynb #t6hak044rc
def suggest_patch_augmentations(patch_size, target_spacing,
                                anisotropy_threshold=3.0,
                                translation_fraction=0.15):
    """Suggest patch-based augmentations with nnU-Net-inspired defaults.

    Derives rotation degrees, translation, and RandomAnisotropy axes from
    patch geometry and voxel spacing. Returns a list of fastMONAI transform
    instances ready for the ``patch_tfms`` parameter in MedPatchDataLoaders.

    Anisotropy detection: if max(spacing)/min(spacing) >= threshold, rotation
    is restricted to 5 deg out-of-plane and 30 deg in-plane. Otherwise 30 deg
    symmetric. Translation is patch_size * fraction per axis.

    Args:
        patch_size: List/tuple of 3 ints -- patch dimensions.
        target_spacing: List/tuple of 3 floats -- voxel spacing.
        anisotropy_threshold: Ratio threshold for anisotropy detection (default 3.0).
        translation_fraction: Fraction of patch_size for translation (default 0.15).

    Returns:
        list: fastMONAI transform instances (7 normally, 6 if RandomAnisotropy omitted).

    Example::

        >>> patch_tfms = suggest_patch_augmentations([128, 128, 32], [0.5, 0.5, 1.5])
        >>> dls = MedPatchDataLoaders.from_df(..., patch_tfms=patch_tfms)
    """
    params = _compute_patch_aug_params(
        patch_size, target_spacing, anisotropy_threshold, translation_fraction
    )
    degrees = params['degrees']
    translation = params['translation']
    aniso_axes = params['aniso_axes']

    # For TorchIO: pass scalar 30 when isotropic (TorchIO expands to symmetric)
    if not params['is_aniso']:
        degrees = 30

    transforms = [
        RandomAffine(scales=(0.7, 1.4), degrees=degrees, translation=translation,
                     default_pad_value=0., p=0.2),
    ]

    if len(aniso_axes) > 0:
        transforms.append(RandomAnisotropy(axes=aniso_axes, downsampling=(1.5, 4), p=0.25))

    transforms.extend([
        RandomGamma(log_gamma=(-0.3, 0.3), p=0.3),
        RandomIntensityScale(scale_range=(0.75, 1.25), p=0.1),
        RandomNoise(std=0.1, p=0.1),
        RandomBlur(std=(0.5, 1.0), p=0.2),
        RandomFlip(p=0.5),
    ])

    return transforms



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/09_external_data.ipynb.

# %% auto #0
__all__ = ['MURLs', 'download_ixi_data', 'download_ixi_tiny', 'download_spine_test_data', 'download_example_spine_data',
           'download_medmnist3d_dataset', 'download_example_endometrial_cancer_data']

# %% ../nbs/09_external_data.ipynb #942e1644-0dac-4701-8a1e-5693f80f3ca0
from pathlib import Path
from glob import glob
from numpy import load 
import pandas as pd
from monai.apps import download_url, download_and_extract
from torchio.datasets.ixi import IXITiny
from torchio import ScalarImage
import multiprocessing as mp
from functools import partial

# %% ../nbs/09_external_data.ipynb #6d5600a3-9d70-4714-8cc2-8f4682a7e5a3
class MURLs():
    """A class with external medical dataset URLs."""

    IXI_DATA = 'http://biomedic.doc.ic.ac.uk/brain-development/downloads/IXI/IXI-T1.tar'
    IXI_DEMOGRAPHIC_INFORMATION = 'http://biomedic.doc.ic.ac.uk/brain-development/downloads/IXI/IXI.xls'
    CHENGWEN_CHU_SPINE_DATA = 'https://drive.google.com/uc?id=1rbm9-KKAexpNm2mC9FsSbfnS8VJaF3Kn&confirm=t'
    EXAMPLE_SPINE_DATA = 'https://drive.google.com/uc?id=1K1J2wmWxYUPY0NbylcV4Otv88EqCL4Ys'
    MEDMNIST_DICT = {'OrganMNIST3D': 'https://zenodo.org/record/6496656/files/organmnist3d.npz?download=1',	
                     'NoduleMNIST3D': 'https://zenodo.org/record/6496656/files/nodulemnist3d.npz?download=1',
                     'AdrenalMNIST3D': 'https://zenodo.org/record/6496656/files/adrenalmnist3d.npz?download=1',	
                     'FractureMNIST3D': 'https://zenodo.org/record/6496656/files/fracturemnist3d.npz?download=1',
                     'VesselMNIST3D': 'https://zenodo.org/record/6496656/files/vesselmnist3d.npz?download=1', 
                     'SynapseMNIST3D': 'https://zenodo.org/record/6496656/files/synapsemnist3d.npz?download=1'}
    EXAMPLE_EC_DATA = 'https://drive.google.com/uc?id=1P_J8MfzbfObTkAyWfY-LlEC_IA_NjYn-'
    EXAMPLE_MULTI_EC_DATA = 'https://drive.google.com/uc?id=1yOmbA9-nKfzzsfysgp9OcPw8GAarXPvE'

# %% ../nbs/09_external_data.ipynb #cb4c061a-d7f5-4ba0-9e7b-ee0c6ec480b0
def _process_ixi_xls(xls_path: (str, Path), img_path: Path) -> pd.DataFrame:
    """Private method to process the demographic information for the IXI dataset.

    Args:
        xls_path: File path to the xls file with the demographic information.
        img_path: Folder path to the images.

    Returns:
        A processed dataframe with image path and demographic information.

    Raises:
        ValueError: If xls_path or img_path do not exist.
    """

    print('Preprocessing ' + str(xls_path))

    df = pd.read_excel(xls_path)

    duplicate_subject_ids = df[df.duplicated(['IXI_ID'], keep=False)].IXI_ID.unique()

    for subject_id in duplicate_subject_ids:
        age = df.loc[df.IXI_ID == subject_id].AGE.nunique()
        if age != 1: df = df.loc[df.IXI_ID != subject_id]  # Remove duplicates with two different age values

    df = df.drop_duplicates(subset='IXI_ID', keep='first').reset_index(drop=True)

    df['subject_id'] = ['IXI' + str(subject_id).zfill(3) for subject_id in df.IXI_ID.values]
    df = df.rename(columns={'SEX_ID (1=m, 2=f)': 'gender'})
    df['age_at_scan'] = df.AGE.round(2)
    df = df.replace({'gender': {1: 'M', 2: 'F'}})

    img_list = list(img_path.glob('*.nii.gz'))
    for path in img_list:
        subject_id = path.parts[-1].split('-')[0]
        df.loc[df.subject_id == subject_id, 't1_path'] = str(path)

    df = df.dropna()
    df = df[['t1_path', 'subject_id', 'gender', 'age_at_scan']]
    
    return df

# %% ../nbs/09_external_data.ipynb #6714a68f-1378-46b3-aeff-ef940213ac2f
def download_ixi_data(path: (str, Path) = '../data') -> Path:
    """Download T1 scans and demographic information from the IXI dataset.
    
    Args:
        path: Path to the directory where the data will be stored. Defaults to '../data'.

    Returns:
        The path to the stored CSV file.
    """

    path = Path(path) / 'IXI'
    img_path = path / 'T1_images'

    # Check whether image data already present in img_path:
    is_extracted = False
    try:
        if len(list(img_path.iterdir())) >= 581:  # 581 imgs in the IXI dataset
            is_extracted = True
            print(f"Images already downloaded and extracted to {img_path}")
    except (FileNotFoundError, StopIteration, OSError):
        is_extracted = False

    if not is_extracted:
        download_and_extract(url=MURLs.IXI_DATA, filepath=path / 'IXI-T1.tar', output_dir=img_path)
        (path / 'IXI-T1.tar').unlink()

    download_url(url=MURLs.IXI_DEMOGRAPHIC_INFORMATION, filepath=path / 'IXI.xls')

    processed_df = _process_ixi_xls(xls_path=path / 'IXI.xls', img_path=img_path)
    processed_df.to_csv(path / 'dataset.csv', index=False)

    return path

# %% ../nbs/09_external_data.ipynb #7753da8a-93e8-4bf3-8f78-bb158b4280d0
def download_ixi_tiny(path: (str, Path) = '../data') -> Path:
    """Download the tiny version of the IXI dataset provided by TorchIO.

    Args:
        path: The directory where the data will be 
            stored. If not provided, defaults to '../data'.

    Returns:
        The path to the directory where the data is stored.
    """
    
    path = Path(path) / 'IXITiny'
    
    IXITiny(root=str(path), download=True)
    download_url(url=MURLs.IXI_DEMOGRAPHIC_INFORMATION, filepath=path/'IXI.xls')
    
    processed_df = _process_ixi_xls(xls_path=path/'IXI.xls', img_path=path/'image')
    processed_df['labels'] = processed_df['t1_path'].str.replace('image','label')
    
    processed_df.to_csv(path/'dataset.csv', index=False)
    
    return path

# %% ../nbs/09_external_data.ipynb #b466174a-4b49-4a8f-92c6-1e5e3ca9fc2a
def _create_spine_df(dir: Path) -> pd.DataFrame:
    """Create a pandas DataFrame containing information about spinal images.

    Args:
        dir: Directory path where data (image and segmentation 
            mask files) are stored.

    Returns:
         A DataFrame containing the paths to the image files and their 
            corresponding mask files, the subject IDs, and a flag indicating that 
            these are test data.
    """
    
    img_list = glob(str(dir / 'img/*.nii.gz'))
    mask_list = [str(fn).replace('img', 'seg') for fn in img_list]
    subject_id_list = [fn.split('_')[-1].split('.')[0] for fn in mask_list]
    
    test_data = {
        't2_img_path': img_list,
        't2_mask_path': mask_list,
        'subject_id': subject_id_list,
        'is_test': True,
    }

    return pd.DataFrame(test_data)

# %% ../nbs/09_external_data.ipynb #26256dca-d9df-43f6-b8eb-f36ce2a445dc
def download_spine_test_data(path: (str, Path) = '../data') -> pd.DataFrame:
    """Downloads T2w scans from the study 'Fully Automatic Localization and 
    Segmentation of 3D Vertebral Bodies from CT/MR Images via a Learning-Based 
    Method' by Chu et. al. 

    Args:
        path: Directory where the downloaded data 
            will be stored and extracted. Defaults to '../data'.

    Returns:
        Processed dataframe containing image paths, label paths, and subject IDs.
    """
    
    study = 'chengwen_chu_2015'
    
    download_and_extract(
        url=MURLs.CHENGWEN_CHU_SPINE_DATA, 
        filepath=f'{study}.zip', 
        output_dir=path
    )
    Path(f'{study}.zip').unlink()
    
    return _create_spine_df(Path(path) / study)

# %% ../nbs/09_external_data.ipynb #93b77ec9-a93a-42cc-b707-4e1e75063533
def download_example_spine_data(path: (str, Path) = '../data') -> Path:
    """Downloads example T2w scan and corresponding predicted mask.
    
    Args:
        path: Directory where the downloaded data 
            will be stored and extracted. Defaults to '../data'.

    Returns:
        Path to the directory where the example data has been extracted.
    """
    
    study = 'spine_example_data'
    
    download_and_extract(
        url=MURLs.EXAMPLE_SPINE_DATA, 
        filepath=f'{study}.zip', 
        output_dir=path
    )
    Path(f'{study}.zip').unlink()
    
    return Path(path) / study

# %% ../nbs/09_external_data.ipynb #89f60f76-5e5d-4a40-b49a-617ec2d35731
def _process_medmnist_img(path, idx_arr):
    """Save tensor as NIfTI."""
    
    idx, arr = idx_arr
    img = ScalarImage(tensor=arr[None, :])
    fn = path/f'{idx}_nodule.nii.gz'
    img.save(fn)
    return str(fn)

# %% ../nbs/09_external_data.ipynb #b0199679-707b-445d-8467-3d342246322a
def _df_sort_and_add_columns(df, label_list, is_val):
    """Sort the dataframe based on img_idx and add labels and if it is validation data column."""
    
    df = df.sort_values(by='img_idx').reset_index(drop=True)
    df['labels'], df['is_val'] = label_list, is_val     
    #df = df.replace({"labels": {0:'b', 1:'m'}})
    df = df.drop('img_idx', axis=1)
    
    return df 

# %% ../nbs/09_external_data.ipynb #d82157b8-ab69-4a38-9323-af4b58c6b54a
def _create_nodule_df(pool, output_dir, imgs, labels, is_val=False): 
    """Create dataframe for MedMNIST data."""
    
    img_path_list = pool.map(partial(_process_medmnist_img, output_dir), enumerate(imgs))
    img_idx = [float(Path(fn).parts[-1].split('_')[0]) for fn in img_path_list]
    
    df = pd.DataFrame(list(zip(img_path_list, img_idx)), columns=['img_path','img_idx'])        
    return  _df_sort_and_add_columns(df, labels, is_val)

# %% ../nbs/09_external_data.ipynb #04095111-a38f-4d00-a0fb-2dbfac263eb6
def _dataset_available_locally(path: Path):
    """Checks if the dataset exists in the local directory."""
    
    return path.exists() and any(path.iterdir())

# %% ../nbs/09_external_data.ipynb #a1b456f0-8e14-42c6-a1cd-a332312145ff
def download_medmnist3d_dataset(study: str, path: (str, Path) = '../data',
                                max_workers: int = 1, skip_download: bool = True):
    """Downloads and processes a particular MedMNIST3D dataset.

    Args:
        study: MedMNIST dataset ('OrganMNIST3D', 'NoduleMNIST3D', 
               'AdrenalMNIST3D', 'FractureMNIST3D', 'VesselMNIST3D', 'SynapseMNIST3D')
        path: Directory to store and extract downloaded data. Defaults to '../data'.
        max_workers: Maximum number of worker processes for data processing. 
                     Defaults to 1.
        skip_download: Flag to indicate if the dataset should be re-downloaded if the dataset exists locally. 
                  Default set to False.

    Returns:
        Two pandas DataFrames. The first DataFrame combines training and validation 
        data, and the second DataFrame contains the testing data.
    """
                                    
    path = Path(path) / study                           
    if _dataset_available_locally(path) and skip_download: 
        return pd.read_csv(path / 'train_val.csv'), pd.read_csv(path / 'test.csv')                                     
                                    
    dataset_file_path = path / f'{study}.npz'

    try:
        download_url(url=MURLs.MEDMNIST_DICT[study], filepath=dataset_file_path)
    except KeyError:
        raise ValueError(f"Dataset '{study}' does not exist.")

    data = load(dataset_file_path)
    keys = ['train_images', 'val_images', 'test_images']

    for key in keys:
        (path / key).mkdir(exist_ok=True)

    train_imgs = data[keys[0]]
    val_imgs = data[keys[1]]
    test_imgs = data[keys[2]]

    with mp.Pool(processes=max_workers) as pool:
        train_df = _create_nodule_df(pool, path / keys[0], train_imgs, 
                                     data['train_labels'])
        val_df = _create_nodule_df(pool, path / keys[1], val_imgs, 
                                   data['val_labels'], is_val=True)
        test_df = _create_nodule_df(pool, path / keys[2], test_imgs, 
                                    data['test_labels'])

    train_val_df = pd.concat([train_df, val_df], ignore_index=True)

    dataset_file_path.unlink()

    train_val_df.to_csv(path / 'train_val.csv', index=False)
    test_df.to_csv(path / 'test.csv', index=False)

    return train_val_df, test_df

# %% ../nbs/09_external_data.ipynb #67491c4a-8491-4b10-8f03-87ffa012fee0
def download_example_endometrial_cancer_data(path: (str, Path) = '../data', multi_channel: bool = True):
    """Download example data for endometrial cancer.
    
    Args:
        path (str or Path): Path to save the downloaded data.
        multi_channel (bool): If True, download multi-channel data (Vibe, T2, 
            and ADC). Otherwise, download only Vibe.
    """
    
    download_url = (MURLs.EXAMPLE_MULTI_EC_DATA if multi_channel 
                    else MURLs.EXAMPLE_EC_DATA)
    
    download_and_extract(
        url=download_url, 
        filepath='ec_example_data.zip', 
        output_dir=path)
    
    Path('ec_example_data.zip').unlink()



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/00_vision_plot.ipynb.

# %% auto #0
__all__ = ['validate_anatomical_plane', 'show_med_img', 'find_max_slice', 'show_segmentation_comparison']

# %% ../nbs/00_vision_plot.ipynb #7ea97ee7-8a55-485d-a75c-ae1a8055f489
import warnings
from fastai.data.all import *
from torchio.visualization import rotate

# %% ../nbs/00_vision_plot.ipynb #55a13906-aef5-445a-8919-eb7daa3816ea
def _get_slice(image, channel: int, slice_index: int, anatomical_plane: int, voxel_size: (int, list)):
    """
    A private method to get a 2D tensor and aspect ratio for plotting.
    This is modified code from the torchio function `plot_volume`.

    Args:
        image: The input image.
        channel: Channel of the image.
        slice_index: Index of the 2D slice.
        anatomical_plane: Anatomical plane of the image.
        voxel_size: Voxel size for the image.

    Returns:
        A 2D tensor containing the sliced image and its aspect.
    """
    if voxel_size is None:
        warnings.warn('Voxel size not defined. Aspect ratio of the plot might not be correct.')
        voxel_size = 1

    if not is_listy(voxel_size):
        voxel_size = (voxel_size, voxel_size, voxel_size)

    sr, sa, ss = voxel_size
    data = image.data[channel]

    if slice_index is None:
        slice_center = np.array(data.shape) // 2
        slice_index = slice_center[anatomical_plane]

    operations = {
        0: (lambda d, i: d[i, :, :], ss / sa),
        1: (lambda d, i: d[:, i, :], ss / sr),
        2: (lambda d, i: d[:, :, i], sa / sr)
    }

    sliced_function, aspect_ratio = operations.get(anatomical_plane)
    sliced_img = rotate(sliced_function(data, slice_index), radiological=True, n=1)

    return sliced_img, aspect_ratio

# %% ../nbs/00_vision_plot.ipynb #4a0a8991-fbb8-4d07-80b6-804852139253
def validate_anatomical_plane(anatomical_plane):
    """Ensure anatomical_plane is either 0, 1, or 2."""
    if anatomical_plane not in [0, 1, 2]:
        raise ValueError("anatomical_plane should be either 0, 1, or 2.")

# %% ../nbs/00_vision_plot.ipynb #7955e15b-7580-4219-838e-93ff094e146a
@delegates(plt.Axes.imshow, keep=True, but=['shape', 'imlim'])
def show_med_img(im, ctx, channel: int, slice_index: int, anatomical_plane: int,
                 voxel_size: (int, list), ax=None, figsize=None, title=None, **kwargs):
    """Show an image on `ax`. This is a modified code from the fastai function `show_image`.

    Args:
        im: The input image.
        ctx: The context.
        channel: Channel of the image.
        slice_index: Index of the 2D slice.
        anatomical_plane: Anatomical plane of the image.
        voxel_size: Voxel size for the image.
        ax: Axis for the plot.
        figsize: Figure size for the plot.
        title: Title for the plot.
        kwargs: Additional parameters for plt.Axes.imshow method.

    Returns:
        Axis with the plot.
    """
                     
    validate_anatomical_plane(anatomical_plane)
                     
    if hasattrs(im, ('data', 'cpu', 'permute')): # Check if `im` has the necessary attributes
        im = im.data.cpu()
        im, aspect = _get_slice(
            im, 
            channel=channel, 
            anatomical_plane=anatomical_plane,
            voxel_size=voxel_size, 
            slice_index=slice_index
        )

    ax = ax if ax is not None else ctx 

    if ax is None: # ax is only None when .show() is used.
        _, ax = plt.subplots(figsize=figsize)

    ax.imshow(im, aspect=aspect, **kwargs)

    if title is not None:
        ax.set_title(title)

    ax.axis('off')

    return ax

# %% ../nbs/00_vision_plot.ipynb #c31d23c6-d819-4039-8d53-ea36ef20f049
def find_max_slice(mask_data, anatomical_plane):
    """Find slice index based on mask"""
    
    axis_mapping = {0: (1, 2),
                    1: (0, 2),
                    2: (0, 1)}

    sums = mask_data.sum(axis=axis_mapping.get(anatomical_plane, (1, 2)))
    idx = np.argmax(sums)

    return idx

# %% ../nbs/00_vision_plot.ipynb #ah52hmsl0of
def show_segmentation_comparison(
    image, ground_truth, prediction,
    slice_index: int = None,
    anatomical_plane: int = 2,
    metric_value: float = None,
    metric_name: str = 'DSC',
    channel: int = 0,
    figsize: tuple = (15, 5),
    cmap_img: str = 'gray',
    cmap_mask: str = 'gray',
    voxel_size = None
):
    """Display 3-panel comparison: Image | Ground Truth | Prediction.
    
    Useful for validating segmentation results, especially after patch-based 
    inference where results are not in standard fastai batch format.
    
    Args:
        image: Input image (MedImage, MedMask, or tensor [C, H, W, D])
        ground_truth: Ground truth mask (MedMask or tensor [C, H, W, D])
        prediction: Predicted mask (tensor [C, H, W, D])
        slice_index: Slice to display. If None, uses find_max_slice on ground_truth.
        anatomical_plane: 0=sagittal, 1=coronal, 2=axial (default)
        metric_value: Optional metric value to display in prediction title
        metric_name: Name of metric for title (default 'DSC')
        channel: Channel to display for multi-channel data (default 0)
        figsize: Figure size (default (15, 5))
        cmap_img: Colormap for image (default 'gray')
        cmap_mask: Colormap for masks (default 'gray')
        voxel_size: Voxel spacing for aspect ratio. If None, aspect=1.

    Example::
    
        # After patch_inference()
        show_segmentation_comparison(
            image=val_img,
            ground_truth=val_gt,
            prediction=predictions[0],
            metric_value=results_df.iloc[0]['DSC'],
            anatomical_plane=2
        )
    """
    validate_anatomical_plane(anatomical_plane)
    
    # Ensure we have tensor data
    img_data = image.data if hasattr(image, 'data') else image
    gt_data = ground_truth.data if hasattr(ground_truth, 'data') else ground_truth
    pred_data = prediction.data if hasattr(prediction, 'data') else prediction
    
    # Move to CPU if needed
    if hasattr(img_data, 'cpu'): img_data = img_data.cpu()
    if hasattr(gt_data, 'cpu'): gt_data = gt_data.cpu()
    if hasattr(pred_data, 'cpu'): pred_data = pred_data.cpu()
    
    # Find optimal slice if not provided
    if slice_index is None:
        gt_np = gt_data[channel].numpy()
        slice_index = find_max_slice(gt_np, anatomical_plane)
    
    # Create figure
    fig, axes = plt.subplots(1, 3, figsize=figsize)
    
    # Get slices with proper aspect ratio using existing _get_slice helper
    img_slice, img_aspect = _get_slice(img_data, channel, slice_index, anatomical_plane, voxel_size)
    gt_slice, gt_aspect = _get_slice(gt_data, channel, slice_index, anatomical_plane, voxel_size)
    pred_slice, pred_aspect = _get_slice(pred_data, channel, slice_index, anatomical_plane, voxel_size)
    
    # Plot panels
    axes[0].imshow(img_slice, cmap=cmap_img, aspect=img_aspect)
    axes[0].set_title('Input Image')
    axes[0].axis('off')
    
    axes[1].imshow(gt_slice, cmap=cmap_mask, aspect=gt_aspect)
    axes[1].set_title('Ground Truth')
    axes[1].axis('off')
    
    # Build prediction title
    pred_title = 'Prediction'
    if metric_value is not None:
        pred_title = f'Prediction ({metric_name}: {metric_value:.4f})'
    
    axes[2].imshow(pred_slice, cmap=cmap_mask, aspect=pred_aspect)
    axes[2].set_title(pred_title)
    axes[2].axis('off')
    
    plt.tight_layout()



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/06_vision_inference.ipynb.

# %% auto #0
__all__ = ['save_series_pred', 'load_system_resources', 'inference', 'compute_binary_tumor_volume', 'refine_binary_pred_mask',
           'keep_largest', 'gradio_image_classifier']

# %% ../nbs/06_vision_inference.ipynb #f28504fd-0e10-4a0e-8ac5-3be955aa3622
from copy import copy
from pathlib import Path
import torch
import numpy as np
from scipy.ndimage import label
from skimage.morphology import remove_small_objects
from SimpleITK import DICOMOrient, GetArrayFromImage
from torchio import Resize, Image
from .vision_core import *
from .vision_augmentation import do_pad_or_crop
from .utils import load_variables
from imagedata.series import Series
from fastai.learner import load_learner

# %% ../nbs/06_vision_inference.ipynb #a3d8ecaf-d80c-463b-84f8-197e8966b304
def _update_uid(attribute, series_obj, val='1234', slice_idx=None):
    """Updates a DICOM UID by replacing its last 4 characters with the provided value."""
    
    uid = series_obj.getDicomAttribute(attribute, slice=slice_idx)[:-4] + val
    series_obj.setDicomAttribute(attribute, uid, slice=slice_idx)
    return series_obj

# %% ../nbs/06_vision_inference.ipynb #4bd80eb1-915e-4ea1-91e0-b68738743e79
def save_series_pred(series_obj, save_dir, val='1234'):
    """Saves series prediction with updated DICOM UIDs."""
    
    series_obj.seriesInstanceUID = series_obj.seriesInstanceUID[:-4] + val
    
    for slice_idx in range(series_obj.slices):
        series_obj = _update_uid('SOPInstanceUID', series_obj, val, slice_idx)
        series_obj = _update_uid('SeriesInstanceUID', series_obj, val, slice_idx)
        
    series_obj.write(save_dir, opts={'keep_uid': True}, formats=['dicom'])

# %% ../nbs/06_vision_inference.ipynb #e4a2771e-25aa-4669-bda2-abcbe29d0504
def _to_original_orientation(input_img, org_orientation):
    """Reorients the image to its original orientation."""
    
    orientation_itk = DICOMOrient(input_img, org_orientation)
    reoriented_array =  GetArrayFromImage(orientation_itk).transpose()
    
    return reoriented_array[None]

# %% ../nbs/06_vision_inference.ipynb #75a1169f-7385-4c48-9a24-51994c80732c
def _do_resize(o, target_shape, image_interpolation='linear', 
               label_interpolation='nearest'):
    """Resample images so the output shape matches the given target shape."""

    resize = Resize(
        target_shape, 
        image_interpolation=image_interpolation, 
        label_interpolation=label_interpolation
    )
    
    return resize(o)

# %% ../nbs/06_vision_inference.ipynb #ee89fef9-1e9d-4d70-8c93-e44e2d600c70
def load_system_resources(models_path, learner_fn, variables_fn):
    """Load necessary resources like learner and variables."""

    learn = load_learner(models_path / learner_fn, cpu=True) 
    vars_fn = models_path / variables_fn
    _, apply_reorder, target_spacing = load_variables(pkl_fn=vars_fn)

    return learn, apply_reorder, target_spacing

# %% ../nbs/06_vision_inference.ipynb #2ba94a55-2cd1-4fae-b8b4-e82051be0cdc
def inference(learn_inf, apply_reorder, target_spacing, fn: (str, Path) = '',
              save_path: (str, Path) = None, org_img=None, input_img=None,
              org_size=None): 
    """Predict on new data using exported model."""         
    
    if None in [org_img, input_img, org_size]: 
        org_img, input_img, org_size = med_img_reader(fn, apply_reorder, target_spacing, 
                                                      only_tensor=False)
    else: 
        org_img, input_img = copy(org_img), copy(input_img)
    
    pred, *_ = learn_inf.predict(input_img.data)
    
    pred_mask = do_pad_or_crop(pred.float(), input_img.shape[1:], padding_mode=0, 
                               mask_name=None)
    input_img.set_data(pred_mask)
    
    input_img = _do_resize(input_img, org_size, image_interpolation='nearest')
    
    reoriented_array = _to_original_orientation(input_img.as_sitk(), 
                                                ('').join(org_img.orientation))
    
    # Create new TorchIO image with proper affine matrix
    # Spatial properties (spacing, origin, direction) are automatically derived from affine
    from torchio import ScalarImage
    import numpy as np
    
    # Ensure we have a valid affine matrix
    if hasattr(org_img, 'affine') and org_img.affine is not None:
        affine_matrix = org_img.affine.copy()
    elif MedBase.affine_matrix is not None:
        affine_matrix = MedBase.affine_matrix.copy()
    else:
        # Fallback to identity affine if no affine available
        affine_matrix = np.eye(4, dtype=np.float64)
    
    pred_img = ScalarImage(tensor=reoriented_array, affine=affine_matrix)

    if save_path:
        save_fn = Path(save_path)/('pred_' + Path(fn).parts[-1])
        pred_img.save(save_fn)
        return save_fn
    
    return pred_img

# %% ../nbs/06_vision_inference.ipynb #1730df1e-e6d7-4fa7-b40e-83f4dda26dc5
def compute_binary_tumor_volume(mask_data: Image):
    """Compute the volume of the tumor in milliliters (ml)."""
    
    dx, dy, dz = mask_data.spacing
    voxel_volume_ml = dx * dy * dz / 1000  
    return np.sum(mask_data) * voxel_volume_ml

# %% ../nbs/06_vision_inference.ipynb #a64a3407-4b97-4b1c-933c-d4a316dbff94
def refine_binary_pred_mask(pred_mask, 
                            remove_size: (int, float) = None,
                            percentage: float = 0.2,
                            verbose: bool = False) -> torch.Tensor:
    """Removes small objects from the predicted binary mask.

    Args:
        pred_mask: The predicted mask from which small objects are to be removed.
        remove_size: The size under which objects are considered 'small'.
        percentage: The percentage of the remove_size to be used as threshold. 
            Defaults to 0.2.
        verbose: If True, print the number of components. Defaults to False.

    Returns:
        The processed mask with small objects removed.
    """
                                
    labeled_mask, n_components = label(pred_mask)

    if verbose:
        print(n_components)

    # Handle empty mask case (no foreground components)
    if n_components == 0:
        return torch.zeros_like(torch.Tensor(pred_mask)).float()

    if remove_size is None:
        sizes = np.bincount(labeled_mask.ravel())
        max_label = sizes[1:].argmax() + 1
        remove_size = sizes[max_label]

    small_objects_threshold = remove_size * percentage
    processed_mask = remove_small_objects(
        labeled_mask, min_size=small_objects_threshold)

    return torch.Tensor(processed_mask > 0).float()

# %% ../nbs/06_vision_inference.ipynb #0frrjbq62loa
def keep_largest(pred_mask: torch.Tensor) -> torch.Tensor:
    """Keep only the largest connected component in a binary mask.

    Args:
        pred_mask: Binary prediction mask tensor.

    Returns:
        Binary mask with only the largest connected component.
    """
    mask_np = pred_mask.numpy() if isinstance(pred_mask, torch.Tensor) else pred_mask
    labeled_mask, n_components = label(mask_np)

    if n_components == 0:
        return torch.zeros_like(pred_mask) if isinstance(pred_mask, torch.Tensor) else mask_np

    sizes = np.bincount(labeled_mask.ravel())
    largest_label = sizes[1:].argmax() + 1  # Skip background (label 0)

    result = (labeled_mask == largest_label).astype(np.float32)
    return torch.from_numpy(result) if isinstance(pred_mask, torch.Tensor) else result

# %% ../nbs/06_vision_inference.ipynb #c2b11267-d760-44b0-bd1b-06fd392ce911
def gradio_image_classifier(file_obj, learn, apply_reorder, target_spacing):
    """Predict on images using exported learner and return the result as a dictionary."""
    
    img_path = Path(file_obj.name)
    img = med_img_reader(img_path, apply_reorder=apply_reorder, target_spacing=target_spacing)
    
    _, _, predictions = learn.predict(img)
    prediction_dict = {index: value.item() for index, value in enumerate(predictions)}

    return prediction_dict



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/04_vision_loss_functions.ipynb.

# %% auto #0
__all__ = ['CustomLoss', 'TverskyFocalLoss']

# %% ../nbs/04_vision_loss_functions.ipynb #e8eee6b3-75c3-4468-8645-8f6a42a76bfe
import torch
from .vision_core import *
from .vision_data import pred_to_binary_mask, batch_pred_to_multiclass_mask
from monai.losses import TverskyLoss, FocalLoss
from torch.nn.modules.loss import _Loss

# %% ../nbs/04_vision_loss_functions.ipynb #e0c0c220-aaeb-46c6-8d18-f72cd9da0555
class CustomLoss:
    """A custom loss wrapper class for loss functions to allow them to work with
    the 'show_results' method in fastai. 
    """

    def __init__(self, loss_func):
        """Constructs CustomLoss object."""
        
        self.loss_func = loss_func

    def __call__(self, pred, targ):
        """Computes the loss for given predictions and targets."""
        
        if isinstance(pred, MedBase):
            pred, targ = torch.Tensor(pred.cpu()), torch.Tensor(targ.cpu().float())
            
        return self.loss_func(pred, targ)

    def activation(self, x):
        return x
    
    def decodes(self, x) -> torch.Tensor:
        """Converts model output to target format.
        
        Args:
            x: Activations for each class with dimensions [B, C, W, H, D].

        Returns:
            The predicted mask.
        """
        
        n_classes = x.shape[1]
        if n_classes == 1: 
            x = pred_to_binary_mask(x)
        else: 
            x,_ = batch_pred_to_multiclass_mask(x)

        return x

# %% ../nbs/04_vision_loss_functions.ipynb #5052c7bc-3d9a-4e34-8b64-bceaf2fdc7b6
class TverskyFocalLoss(_Loss):
    """
    Compute Tversky loss with a focus parameter, gamma, applied.
    The details of Tversky loss is shown in ``monai.losses.TverskyLoss``.
    """

    def __init__(
        self,
        include_background: bool = True,
        to_onehot_y: bool = False,
        sigmoid: bool = False,
        softmax: bool = False,
        gamma: float = 2,
        alpha: float = 0.5, 
        beta: float = 0.99):
        """
        Args:
            include_background: if to calculate loss for the background class.
            to_onehot_y: whether to convert `y` into one-hot format.
            sigmoid: if True, apply a sigmoid function to the prediction.
            softmax: if True, apply a softmax function to the prediction.
            gamma: the focal parameter, it modulates the loss with regards to 
                how far the prediction is from target.
            alpha: the weight of false positive in Tversky loss calculation.
            beta: the weight of false negative in Tversky loss calculation.
        """
            
        super().__init__()
        self.tversky = TverskyLoss(
            to_onehot_y=to_onehot_y, 
            include_background=include_background, 
            sigmoid=sigmoid, 
            softmax=softmax, 
            alpha=alpha, 
            beta=beta
        )
        self.gamma = gamma

    def forward(self, input: torch.Tensor, target: torch.Tensor) -> torch.Tensor:
        """
        Args:
            input: the shape should be [B, C, W, H, D]. The input should be the original logits.
            target: the shape should be[B, C, W, H, D].

        Raises:
            ValueError: When number of dimensions for input and target are different.
        """
        if len(input.shape) != len(target.shape):
            raise ValueError("The number of dimensions for input and target should be the same.")

        tversky_loss = self.tversky(input, target)
        total_loss: torch.Tensor = 1 - ((1 - tversky_loss)**self.gamma)

        return total_loss
